/* -*- c++ -*- */
/*
 * Copyright 2016 The Aerospace Corporation.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */
/* GOES Radio Comments
GOESProcessor.cpp is the class that does the data link layer processing for EMWIN-N
and HRIT/LRIT.
InitializeProcessor(), Initializes the memory buffers
UninitializeProcessor(), Deletes the memory buffers

ProcessLritDataBLock(Buffer, Buffer size, x,y) is a wrapper for LRIT_main(x,y).  It
takes the data from buffer, and puts it into a global array, CurrentDataPtr,  that
LRIT_main can access.  This global array was initialized in InitializeProcessor()

LRIT_main(resynch threshold, BERFIleNAme) takes the data and performs frame synchronization.
If there is a 180 degree phase ambiguity, it resolves it by correlating to the inverted
header sequence.  LRIT_main uses a state-machine with 4 levels of lock (0 to 3) to maintain frame
synchronization.  the user-defined resynch threshold is the threshold for lock-state 3 (total lock).
After Frame Synch, LRIT_main does Reed Solomon Decoding, LRITrs->Decode().  The RS-decoded
CADU frame data is stored
in LRITframe.  From here, the function can write status messages like BER and FER to a file
(BERFileName) and to the screen.  It reads
the VCDU. It also reads the APIDs using GetLritPacketHeader();  During this function call, LRITFile[]
gets assigned to LritFileptr[].  When GetLritPacketHeader returns, it copies the processed LRIT data
(LRITframe) into an 2-D array LritFilePtr[APIDindex].  LritFilePtr[0] would represent a pointer to one
LRIT data frame.  We store multiple file parts at a time, and index each one based on the APIDindex
table.  Because LritFilePtr[] points to LRITFile[], LRITFile[] also gets the contents of the LRITframe(s).
Frome here it calls PRocessLRITFile

ProcessLRITFile()
This function associates a new global array, LpBufPtr to LRITFile[APIDindex],
thus getting access to a single LRIT frame data.  It eventually writes the actual LRIT data products
(Saved in another global array called FileBuffer) to file. Before it does this, signficant processing
is done in CalcCRC().

CalcCRC()
CalcCRC grabs data from the global array lpBufPtr.  It also associates a new global array, filepointer,
to FileBuffer.  FileBuffer is used by ProcessLRITFile() to write the LRIT products to disk. Using
LpBufPtr, it checks validates the CRC and copies the contents of LpBufPtr into filepointer (which is
pointing to FileBuffer). It evaluates the contents of the LRIT data (LpBufPtr).  If its header data,
it parses it for the following information:  Primary Header, Header Structure Record, NOAA LRIT Record,
Image Structure Record,Image Structure Record, Image Navigation Record, Image Data Function,
Annotation Record, Time Stamp, Ancillary Text Record, Image Segment ID Record, Rice Compression Record,
Data Filename, and EMWIN file marker
when it gets to the body of the data, based on the header, it either performs rice decompression or
sends the emwin data out to a emwin processor.  For the Rice Decompression, it copies copies the
contents of LpBufPtr to riceptr, which is a global array that RiceDecompress() uses.  If
RiceDecompress is successful, its output is copied into LRITImageptr.  Finalyl, LRITImageptr is
copied into filepoint, which points to (FileBuffer).  CalcCRC() exits, and ProcessLRITFile() writes
the uncompressed LRIT data to a file.  If the packet contains EMWIN data, it copies the contents of
LpBufPtr into a global array called EmwinPtr and then calls ProcessEmwinFile().



For EMWIN-N processing
ProcessEmwinDataBLock(Buffer, Buffer size, x,y) is called.
It takes the data from buffer, and puts it into a global array, CurrentDataPtr, that EMWIN_main() can
access.  This global array was initialized in InitializeProcessor().

EMWIN_main(resynch threshold, BERFIleNAme) and LRIT_main() are identical, until after RS Solomon
Decoding is performed.  The buffer that contains the decoded EMWIN-N CADU frame data is called LRITframe
unfortunately due to copy/paste issues.  The APIDS for EMWIN-N are not read like they are for LRIT. The
VCDU header is checked to see if the current EMWIN frame is a fill or data packet.  If its a valid data
packet, LRITframe is copied to a global array EmwinPtr, and then ProcessEmwinFile() is called.

ProcessEmwinFile()
EmwinPtr is linked to the global array EmwinProcPtr.  EmwinProcPtr is evaluated for the six nulls and the
file name.  There are checks throughout the code that makes sure there is enough data in the temporary
buffers to finish processing.  Once the file name is found, the emwin packet is stored in a packet array
that stores the temporary files via EmwinPtrtabl[][] = EmwinProcPtr.  When enough emwin frames are collected
to complete a file, emwindisplay() is called

Emwindisplay()
Parses the data, and writes zip, txt, gif, and jpeg files to disk.
*/
#include <iostream>
#include <string>
#include <cstdio>
#include <cmath>
#include <cstdlib>
//#include <cstddef>
#include <fstream>
#include <algorithm>

#include <boost/algorithm/string.hpp>
#include <boost/format.hpp>

#if defined __GNUC__
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#endif

#include "GoesProcessor.h"
extern "C" {
#include <szlib.h>
}

#define RE_ACQ                          //Enables ReAcq Support
#define V2_2
#define NO_SYNC_LIMIT     (600*1024)    //ReAcq is engaged after 600 frames are missed
#define MIN_DATA_SIZE        (2048)     //Minimum number of bytes before beginning packet processing
#define FILENAME_SIZE          (12)     //Size of the EMWIN File Names
#define HEADER_SIZE            (80)
#define WAIT_TIL_DISCARD      (100)     //If its been 100 packets since a part of an incomplete file is received, then discard.
#define NUMBER_OF_FILES        (200)     //maximum number of partial files stored in memeory at once
#define TMP_BUFFER_SIZE  (150000000)     //Version Release 1 and 1.29 Stores temporary data. The bigger, the more stable, but the more memory used.
#define BIG_BUFFER_SIZE  (500000000)    //This buffer holds all processed data. The larger it is, the safer, but the more memory used. 150 MB (Moore's Law Factor)


//#define TMP_BUFFER_SIZE  (10000000)     //Version Release 1 and 1.29 Stores temporary data. The bigger, the more stable, but the more memory used.
//#define BIG_BUFFER_SIZE  (150000000)    //This buffer holds all processed data. The larger it is, the safer, but the more memory used. 150 MB (Moore's Law Factor)
#define LRIT_PACKET_SIZE      (1020)    // bytes

CGoesProcessor::CGoesProcessor()
{
    DataBuffer      = NULL;
    //DataBuffer:: points, always to the begining of the allocated memory
    LastProcBytePtr   = NULL;
    //LastProcBytePtr (oldname: dispind) increases by 1020bytes,in LRIT_Main when a "good" block was
    // found. If a header is not found, it increases by 1, and the search for the header
    // restarts
    // After one is synched, LastProcBytePtr points to the last processed RS byte
    old_LastProcBytePtr   = NULL;
    // Keeps the old value of LastProcBytePtr. It is only updated on synched mode

    CurrentDataPtr        = NULL;
    //CurrentDataPtr: As data comes in from GNURadio, it is copied to the "Processing Buffer"
    // In the function, ProcessLRITDataBlock  , we check to see where the difference
    // (CurrentDataPtr - LastProcBytePtr)  > MIN_DATA_SIZE bytres. When this happens, we go ahead and process it.

    lpBufPtr        = NULL;
    LRITFirstsync   = true;
    LRITsyncflag    = false;
    Firstsyncflag   = true;
    foundsyncflag   = false;

    //------------------------------------------------
    //2.3       EMWIN-OQPSK CCSDS Data Format Summary
    //   CCSDS Data Format Summary
    //ASM       Hexadecimal 1ACFFC1D
    Ltreg           = 0xffffffff;
    lritsync        = 0x1acffc1d;
    //------------------------------------------------
    LRITDataInvert  = 0;
    InvByteCount    = 0;
    InvBadFrameCount= 0;
    LRITFrameCount  = 0;
    PendingAPID     = 2047;
    LRITrs          = NULL;
    VCDUcount       = 0;
    Fullheader      = false;
    Extension       = false;
    InvToggleFlag   = false;
    PacketIndex     = 0;
    SrchHeaderFlag  = false;
    LRIThavename    = false;
    LRITfiletype    = 130;
    FileBuffer      = NULL;
    numLRITbytes    = 0;
    haveLRITfiles   = false;
    FrameSyncCnt    = 0;
    EmwinDataSize   = BIG_BUFFER_SIZE; //JR Debug
    EmwinData       = NULL;
    EmwinBytes      = 0;
    EmwinNullCount  = 0;
    haveEMWINfiles  = false;
    EmwinStart      = false;
    EmwinPacketCount= 0;
    EmwinTempBuffer = NULL;
    EmwinProcBytes  = 0;
    viterbi_framerrors = 0;
    RS_framerrors   = 0;
    ErCountCum      = 0;
    strcpy(statestring, "ALL");
    for (int i = 0; i < NUMBER_OF_FILES; i++)
    {
        LRITFile[i]     = NULL;
        EmwinFileName[i]   = NULL;
    }

    //EV: These Variables used to be Static
    // FRom LRIT_Main
    LRITFrameError =0;
    Missed_FrameCount = 0;
    VCDUmark1 =0;  //Marker to see if we missed a VCDU frame
    VCDUmark0 = 0;
    lock_status = -1; //lock status =-1,1,2,3 means completely_unsynched, somewhat_synched, pretty_synched, extremely_synched
    consecutive_good_frames = 0;
    consecutive_frames_missed = 0;
    consecutive_frame_errors = 0;
    this->bad_data_count =0;
    best_offset=0;
    best_offset_ready_signal = 1;
    //subtract_four = 0;
    MissedFrameInitialState = true;

    //Rice Variables
    NC_failed = 0;

    /// EMWIN Variables
    EMWINFrameCount =0;
    EMWINFrameError =0;
    remain=0;

    gr::configure_default_loggers(d_logger, d_debug_logger, "GoesProcessor");
}

bool CGoesProcessor::InitializeProcessor(toggle_function_t toggle_delay,
                                         toggle_function_t toggle_rotate,
                                         int delay_thresh)
{
    //this->img_sink = NULL;
    this->use_delay = true;
    this->toggle_viterbi_delay = toggle_delay;
    this->toggle_rotate = toggle_rotate;
    this->delay_thresh = delay_thresh;
    num_bad = 0;

    #if defined __GNUC__
    //this->keypress_thread = new boost::thread(boost::bind(&CGoesProcessor::CheckKeyReset, this));
    #endif

    this->bad_data_count =0;
    char randm,feedbk;
    unsigned int i, j;
    InvToggleFlag = false;

    LRITrs = new CReedSolomon(8,16,112,11,0,4,0,1);
    if (!LRITrs)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for reed solomon");
        return false;
    }

    DataBuffer = new unsigned char[BIG_BUFFER_SIZE];
    if (!DataBuffer)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for DataBuffer");
        return false;
    }

    lpBufPtr            = DataBuffer;
    LastProcBytePtr     = DataBuffer;
    invresetpointer     = DataBuffer;
    CurrentDataPtr      = DataBuffer;
    old_LastProcBytePtr = DataBuffer;

    EmwinData = new unsigned char[EmwinDataSize];
    if (!EmwinData)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for EmwinData");
        return false;
    }
    EmwinBufEnd  = EmwinData + EmwinDataSize;
    EmwinPtr     = EmwinData;
    EmwinProcPtr = EmwinData;

    EmwinTempBuffer = new unsigned char[TMP_BUFFER_SIZE];
    if (!EmwinTempBuffer)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for EmwinTempBuffer");
        return false;
    }

    FileBuffer = new unsigned char[TMP_BUFFER_SIZE];
    if (!FileBuffer)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for FileBuffer");
        return false;
    }


    for (i = 0; i < NUMBER_OF_FILES; i++)  {
        //---------------- LRIT --------------------------------
        LRITFile[i] = new unsigned char[TMP_BUFFER_SIZE];
        if (!LRITFile[i]){
            GR_LOG_FATAL(d_logger, "Error allocating memory for LRITFile");
            return false;
        }

        LRITFilePtr[i] = LRITFile[i];
        //---------------- EMWIN --------------------------------
        // Last one is a null
        EmwinFileName[i] = new unsigned char[FILENAME_SIZE+1];
        if (!EmwinFileName[i])
        {
            GR_LOG_FATAL(d_logger, "Error allocating memory for EmwinFileName");
            return false;
        }
        //BUG *(EmwinFileName[i] + FILENAME_SIZE+1) = 0;        //NULL to make into string
        // The way it should be
        //      EmwinFileIndex[0,1,2,3,4,5,6,7] = Characters
        //      EmwinFileIndex[8] = "."
        //      EmwinFileIndex[9,10,11] = Extension
        //      EmwinFileIndex[12] = NULL
        // EmwinFileIndex is an array of "int" so be careful here ...
        //EmwinFileIndex[i][FILENAME_SIZE] = 0; //NULL to make into string        // This is how it should be

        EmwinFileIndex[i][0] = 0;                               //mark slot available
        EmwinFileIndex[i][3] = 0;                               //initialize timeout counter ????
        //Added
        EmwinFileIndex[i][1] = 0;
        EmwinFileIndex[i][2] = 0;
    }

    //-----------------------------------------------------------------------------------
    // RANDOMIZER
    // char Random[1020]; = 8160 bits
    //Iniialize Shift Register to all 1's   --> Specified in standard
    randm = (char)0xff;
    // Original POlynomial is :  1 + x3 + x5 + x7 +x8
    for (i = 0; i < 1020; i++)  //1020bytes = 8160 bits in CADU packet
    {
        Random[i] = 0;
        for (j = 0; j <= 7; j++)
        {
            Random[i] = Random[i] << 1;
            if (randm & 0x80) //80h = 1000 0000b
                Random[i]++;

            //Bit-Wise AND between: Fixed shift register(95h) and the state of the
            // feedback register: randm
            feedbk = randm & 0x95;//95h = 1001 0101--> bits 1,3,5,8
            //feedback contains the contents of the registers masked by the polynomial
            //  1 + x3 + x5 + xt +x8 = 95 h
            randm = randm << 1;

            if ((((feedbk & 0x80)^(0x80 & feedbk << 3))^(0x80 & (feedbk << 5)))^(0x80 & (feedbk << 7)))
                randm++;
        }
    }

    for (i = 0; i < 64; i++)
    {
        for (j = 0; j < 9; j++)
        {
            VCIDtable[i][j] = 0;
            if (j == 3)
                VCIDtable[i][j] = 2047; //Set Pending Apid=2047
            if (j == 2)
                VCIDtable[i][j] = 2049; //Set APID index out of range

        }
    }

    for (i = 0; i < MAX_APID_TABLE; i++) //Initialize LRIT variables
    {
        APIDtable[i] = 2040;
        LengthTable[i] = 0;
        filesequence[i] = 0;
        Bytesremaining[i] = 0;
    }

    return true;
}

bool CGoesProcessor::InitializeProcessor()
{
    this->bad_data_count = 0;
    char randm,feedbk;
    unsigned int i, j;
    InvToggleFlag = false;

    LRITrs = new CReedSolomon(8,16,112,11,0,4,0,1);
    if (!LRITrs)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for reed solomon");
        return false;
    }

    DataBuffer = new unsigned char[BIG_BUFFER_SIZE];
    if (!DataBuffer)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for DataBuffer");
        return false;
    }

    lpBufPtr            = DataBuffer;
    LastProcBytePtr     = DataBuffer;
    invresetpointer     = DataBuffer;
    CurrentDataPtr      = DataBuffer;
    old_LastProcBytePtr = DataBuffer;

    EmwinData = new unsigned char[EmwinDataSize];
    if (!EmwinData)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for EmwinData");
        return false;
    }
    EmwinBufEnd  = EmwinData + EmwinDataSize;
    EmwinPtr     = EmwinData;
    EmwinProcPtr = EmwinData;

    EmwinTempBuffer = new unsigned char[TMP_BUFFER_SIZE];
    if (!EmwinTempBuffer)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for EmwinTempBuffer");
        return false;
    }

    FileBuffer = new unsigned char[TMP_BUFFER_SIZE];
    if (!FileBuffer)
    {
        GR_LOG_FATAL(d_logger, "Error allocating memory for FileBuffer");
        return false;
    }


    for (i = 0; i < NUMBER_OF_FILES; i++)  {
        //---------------- LRIT --------------------------------
        LRITFile[i] = new unsigned char[TMP_BUFFER_SIZE];
        if (!LRITFile[i]){
            GR_LOG_FATAL(d_logger, "Error allocating memory for LRITFile");
            return false;
        }

        LRITFilePtr[i] = LRITFile[i];
        //---------------- EMWIN --------------------------------
        // Last one is a null
        EmwinFileName[i] = new unsigned char[FILENAME_SIZE+1];
        if (!EmwinFileName[i])
        {
            GR_LOG_FATAL(d_logger, "Error allocating memory for EmwinFileName");
            return false;
        }
        //BUG *(EmwinFileName[i] + FILENAME_SIZE+1) = 0;        //NULL to make into string
        // The way it should be
        //      EmwinFileIndex[0,1,2,3,4,5,6,7] = Characters
        //      EmwinFileIndex[8] = "."
        //      EmwinFileIndex[9,10,11] = Extension
        //      EmwinFileIndex[12] = NULL
        // EmwinFileIndex is an array of "int" so be careful here ...
        //EmwinFileIndex[i][FILENAME_SIZE] = 0; //NULL to make into string        // This is how it should be

        EmwinFileIndex[i][0] = 0;                               //mark slot available
        EmwinFileIndex[i][3] = 0;                               //initialize timeout counter ????
        //Added
        EmwinFileIndex[i][1] = 0;
        EmwinFileIndex[i][2] = 0;
    }

    //-----------------------------------------------------------------------------------
    // RANDOMIZER
    // char Random[1020]; = 8160 bits
    //Iniialize Shift Register to all 1's   --> Specified in standard
    randm = (char)0xff;
    // Original POlynomial is :  1 + x3 + x5 + x7 +x8
    for (i = 0; i < 1020; i++)  //1020bytes = 8160 bits in CADU packet
    {
        Random[i] = 0;
        for (j = 0; j <= 7; j++)
        {
            Random[i] = Random[i] << 1;
            if (randm & 0x80) //80h = 1000 0000b
                Random[i]++;

            //Bit-Wise AND between: Fixed shift register(95h) and the state of the
            // feedback register: randm
            feedbk = randm & 0x95;//95h = 1001 0101--> bits 1,3,5,8
            //feedback contains the contents of the registers masked by the polynomial
            //  1 + x3 + x5 + xt +x8 = 95 h
            randm = randm << 1;

            if ((((feedbk & 0x80)^(0x80 & feedbk << 3))^(0x80 & (feedbk << 5)))^(0x80 & (feedbk << 7)))
                randm++;
        }
    }

    for (i = 0; i < 64; i++)
    {
        for (j = 0; j < 9; j++)
        {
            VCIDtable[i][j] = 0;
            if (j == 3)
                VCIDtable[i][j] = 2047; //Set Pending Apid=2047
            if (j == 2)
                VCIDtable[i][j] = 2049; //Set APID index out of range

        }
    }

    for (i = 0; i < MAX_APID_TABLE; i++) //Initialize LRIT variables
    {
        APIDtable[i] = 2040;
        LengthTable[i] = 0;
        filesequence[i] = 0;
        Bytesremaining[i] = 0;
    }

    return true;
}

void CGoesProcessor::UnInitializeProcessor()
{

    if (LRITrs)
        delete LRITrs;
    if (DataBuffer)
        delete [] DataBuffer;
    if (EmwinData)
        delete [] EmwinData;
    if  (EmwinTempBuffer)
        delete [] EmwinTempBuffer;
    if (FileBuffer)
        delete [] FileBuffer;
    for (int i = 0; i < NUMBER_OF_FILES; i++)
    {
        if (LRITFile[i])
            delete [] LRITFile[i];
        if (EmwinFileName[i])
            delete [] EmwinFileName[i];
    }
    this->bad_data_count=0;
}

bool CGoesProcessor::ProcessLritDataBlock(unsigned char *ReadBuffer, int next_block_size, int resynch_thresh)
{
    long unsigned int bytes_remaining;

    if ((CurrentDataPtr + next_block_size) > (DataBuffer + BIG_BUFFER_SIZE)){

        GR_LOG_INFO(d_debug_logger, boost::format("CurrentDataPtr  = %p") % CurrentDataPtr);
        GR_LOG_INFO(d_debug_logger, boost::format("next_block_size  = %i") % next_block_size);
        GR_LOG_INFO(d_debug_logger, boost::format("DataBuffer + BIG_BUFFER_SIZE = %p") % (DataBuffer + BIG_BUFFER_SIZE));
        GR_LOG_INFO(d_debug_logger, boost::format("DataBuffer = %p") % DataBuffer);
        GR_LOG_INFO(d_debug_logger, boost::format("BIG_BUFFER_SIZE = %i") % BIG_BUFFER_SIZE);


        //~~~~~~~~~~~~ ORIGINAL METHOD  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        //Step 1: Copy the unprocessed bytes to the begining of the buffer
        bytes_remaining   =  CurrentDataPtr - LastProcBytePtr;
        memcpy(DataBuffer, LastProcBytePtr, bytes_remaining);
        //Step 2: Make the pointer point to the begining of the buffer
        // We assume this information has already been processed
        lpBufPtr            = DataBuffer;
        LastProcBytePtr     = DataBuffer;
        invresetpointer     = DataBuffer;
        old_LastProcBytePtr = DataBuffer;

        CurrentDataPtr      = DataBuffer + bytes_remaining;
        GR_LOG_WARN(d_logger,
                    boost::format("Buffer Full, relocating unprocessed data:"
                                  "Temp.Buffer Size=%i\tTotal Buf.Size=%i\t"
                                  "Bytes Transferred=%lu")
                                  % TMP_BUFFER_SIZE % BIG_BUFFER_SIZE % bytes_remaining);
    }

    memcpy(CurrentDataPtr, ReadBuffer, next_block_size*sizeof(unsigned char));
    CurrentDataPtr += next_block_size;

#ifdef RE_ACQ
    ReAcquire = 0;
    while ((CurrentDataPtr - LastProcBytePtr >= MIN_DATA_SIZE) && (!ReAcquire)) {

        ReAcquire = LRIT_Main(resynch_thresh);
    }
    // Process exits when LRIT_Main() returns with an error
    if (!ReAcquire){
        return true;
    }else{
        return false;
    }
#else

    while (CurrentDataPtr - LastProcBytePtr >= MIN_DATA_SIZE)
        LRIT_Main(resynch_thresh);

    return true;
#endif

}

bool CGoesProcessor::ProcessEmwinDataBlock(unsigned char *ReadBuffer, int next_block_size, int resynch_thresh)
{
    int bytes_remaining;
    if (CurrentDataPtr + next_block_size > DataBuffer + BIG_BUFFER_SIZE)
    {
        GR_LOG_WARN(d_logger, "Buffer Full: Relocating unprocessed data");

        bytes_remaining = CurrentDataPtr - LastProcBytePtr;
        memcpy(DataBuffer, LastProcBytePtr, bytes_remaining);
        lpBufPtr = DataBuffer;
        LastProcBytePtr = DataBuffer;
        invresetpointer = DataBuffer;
        old_LastProcBytePtr = DataBuffer;
        CurrentDataPtr = DataBuffer + bytes_remaining;
    }

    memcpy(CurrentDataPtr, ReadBuffer, next_block_size);
    CurrentDataPtr += next_block_size;

/*#ifdef RE_ACQ
    ReAcquire = 0;
    while ((CurrentDataPtr - LastProcBytePtr >= MIN_DATA_SIZE)&(!ReAcquire)){
        ReAcquire = LRIT_Main(resynch_thresh);
    }
    if (!ReAcquire){
        return true;
    }else{
        return false;
    }
#else*/
    while (CurrentDataPtr - LastProcBytePtr >= MIN_DATA_SIZE)
        EMWIN_Main(resynch_thresh);
    return true;
//#endif
}

int CGoesProcessor::LRIT_Main(int resynch_thresh)
{

    unsigned char   hreg;
    int             offset;
    int             firstheaderpointer;
    bool            head = false;
    int             bytestowrite = 0;
    char            xcor;
    unsigned int    i, j, k, m, n, ercount, ercountmax = 0;
    unsigned long   synk;
    int             syncsum, syncthresh;
    unsigned short  temword;
    FILE            *BER_File;
    int SEARCH_DEPTH = 5;  //how far you can search for a header in bytes, before downgrading lock status
    int CONSECUTIVE_GOOD_THRESHOLD = 3; //How many consecutive frames I need to receiver in order to be in the highest lock state.
    int CONSECUTIVE_BAD_THRESHOLD = 3;

    int resynch_thresh3 = resynch_thresh;  //Found more than CONSECUTIVE THRESHOLD Frames in the Exact positions you expect to find headers
    int resynch_thresh2 = 6; //Had one missed frame or frame in error but looking in areas where you expect to find headers.
    int resynch_thresh1 = 2; //Looking for header in the middle of a packet
    if (resynch_thresh2 > resynch_thresh3)
        resynch_thresh2 = resynch_thresh3;

    //===============================================
    // Eliminate GOTOs
    bool getAPID= false;
    bool BadData= false;

    //===============================================
    if ((bad_data_count % 1024) == 1){
      GR_LOG_DEBUG(d_debug_logger,
                   boost::format("LRIT_Main::Detecting Unsynched Data=%d ksym")
                                % ((int)bad_data_count/1024));
    }
#ifdef V2_2
    if (num_bad > delay_thresh) {
      this->toggle_viterbi_delay();
      num_bad = 0;
    }
#endif
    //Process Every Byte that is stored in memory for processing
    if ((CurrentDataPtr - LastProcBytePtr) >= MIN_DATA_SIZE)
    {
        if (LRITFirstsync)
        {
            // shift in the first 4 bytes of the packet buffer into Ltreg
            for (i = 0; i < 4; i++)
                Ltreg = (Ltreg << 8) | ((*(LastProcBytePtr + i)) ^ LRITDataInvert);
            LastProcBytePtr += 4; // advance packet buffer start by 4 bytes
            LRITFirstsync = false;
        }

        //LRITDataInvert is used to resolve the 180 degree phase ambiguity.
        hreg = *LastProcBytePtr ^ LRITDataInvert;

        //Maintain byte alignment by checking all 8 bits within the byte for the header
        for (offset = 0; offset <= 7; offset++)
        {
            //LRITsync = CADU Header = 32 bits
            //Ltreg is the 32 bit section of LRIT data being compared against the header
            synk = lritsync ^ Ltreg;    // Bit-wise XOR

            // If the header is found, then sync=32 "0" bits
            syncsum = 0;
            for (n = 0; n < 32; n++)
            {
                if (synk & 0x00000001)
                    syncsum++;
                synk = synk >> 1;
            }

            if ((consecutive_frames_missed + consecutive_frame_errors) > CONSECUTIVE_BAD_THRESHOLD){
                if (lock_status != -1) {
                    GR_LOG_INFO(d_debug_logger,
                                 boost::format("Lock Status Changed from %d to %d")
                                               % lock_status % -1);
                }
                lock_status = -1;

            }

            if (lock_status==1)
                syncthresh = resynch_thresh1;
            else if (lock_status==2)
                syncthresh = resynch_thresh2;
            else if (lock_status>=3 && (offset!=best_offset)) //Wrong expected bit alignment within bytesteam
                syncthresh = resynch_thresh2;
            else if (lock_status>=3 && (offset==best_offset)) //Correct expected bit alignment within bytesteam
                syncthresh = resynch_thresh3;
            else if (lock_status<0)     // to sync initially, need all 32 bits in header correct
                syncthresh = 0;

            if (syncsum <= syncthresh)   // if 32-bit header was found
            {
                FrameSyncCnt ++;
                best_offset_ready_signal = 1; //This is to allows you to estimate and print the syncsum for the undetected headers.
                if (consecutive_good_frames >= CONSECUTIVE_GOOD_THRESHOLD){
                    if (lock_status != 3)
                        //printf("Lock Status Changed From %d to %d\n", lock_status, 3);
                        lock_status = 3;  //upgrade lock status directly to 3
                    best_offset = offset;
                }

                InvByteCount = 0;
                InvBadFrameCount = 0;
                invresetpointer = LastProcBytePtr;
                foundsyncflag = true;
                LRITFirstsync = true;
                LRITsyncflag = true;

                if (!Firstsyncflag){
                    processbytes = (int)(LastProcBytePtr - old_LastProcBytePtr);
                    if (processbytes > 1024)
                        LRITsyncflag = false;
                }else{
                    Firstsyncflag = false;
                }
                old_LastProcBytePtr = LastProcBytePtr;

                //De-randomize the data
                for (j = 0; j < LRIT_PACKET_SIZE; j++)
                {
                    LRITframe[j] = (*(LastProcBytePtr + j)) ^ LRITDataInvert;
                    if (offset != 0)
                    {
                        temword = (*(LastProcBytePtr + j)) ^ LRITDataInvert;
                        temword = (temword << 8) | (*(LastProcBytePtr + j + 1)) ^ LRITDataInvert;
                        LRITframe[j] = temword >> (8 - offset);
                    }
                    LRITframe[j] ^= Random[j];
                }

                LastProcBytePtr += 1020;  //CADU packet length is 8160bits = 1020 bytes
                LRITFrameCount++;

                //Make a copy of the CADU frame before Reed Solomon Decoding
                for (k = 0; k < 1024; k++)
                    LRIT1frame[k] = LRITframe[k];

                //RS Decode
                LRITrs->Decode((unsigned char *) &LRITframe[0]);
                viterbi_framerrors = LRITrs->framerrors;
                // there can be 0 - 4 uncorrected frame errors
                RS_framerrors = LRITrs->uncorrected_framerrors;

                //Compare the LRIT Frame Before and after RS decoding to estimate the
                //Error Rate at the Output of Viterbi Decoding
                ercount=0;
                for (k = 0; k < 1020; k++)
                {
                    xcor = LRIT1frame[k] ^ LRITframe[k];
                    for (m = 0; m < 8; m++)
                    {
                        if ((xcor & 0x01) == 0x01) {
                            ercount++;
                            //ErrCount ++;
                        }
                        xcor = xcor >> 1;
                    }
                }
                ErCountCum = ErCountCum + ercount;

                //If there is a Reed Solomon error, update error counters
                //Change Frame Synchronization Status
                // Try and recover the VCDU counter so you can detect a missed frame
                if (RS_framerrors>0)
                {
                    LRITFrameError += RS_framerrors;
                    consecutive_frame_errors++;
                    consecutive_good_frames = 0;
                    if (lock_status==3){ //Strictly downgrade from 3 to 2 only.
                        GR_LOG_INFO(d_debug_logger,
                                    boost::format("RS frame errors: Lock Status Changed from %d to %d")
                                                  % lock_status % 2);
                        //printf("Lock Status Changed From %d to %d\n", 3, 2);
                        lock_status=2;
                    }
                    //printf("Frame Errors!\t Difference between counters: %d\t 0: %d\t 1: %d\t\n",(VCDUmark0 - VCDUmark1), VCDUmark0,VCDUmark1);
                    return 0;
                }
                // 11499792 = 0x00af7910
                if (VCDUmark0 == 11499792) LRITFrameError -= 4; // SUPER HOTFIX GET RID OF THIS LATER /* JD: WTF?!? */
                //The rest of code in [if(syncsun<sync_thresh)] represents a frame synch and a good received frame

                if (lock_status<2){ //If found and synched to good frame, upgrade from 0 or 1, up to 2.
                    //printf("Lock Status Changed From %d to %d\n", lock_status, 2);
                    GR_LOG_INFO(d_debug_logger,
                                boost::format("RS decode success: Lock Status Changed from %d to %d")
                                              % lock_status % 2);
                    lock_status = 2;
                }
                consecutive_frames_missed = 0;
                consecutive_good_frames++;

                //Get the VCID Marker
                //Compare to to last VCID marker to make sure a frame wasn't missed
                VCID        = LRITframe[1] & 0x3f;
                VCDUcount   = LRITframe[2];
                VCDUcount   = (VCDUcount << 8) | LRITframe[3];
                VCDUcount   = (VCDUcount << 8) | LRITframe[4];
                VCDUmark1   = VCDUmark0;
                VCDUmark0   = VCDUcount;

                //IF the VCDU markers do not agree, declare a missed frame.
                //If there is a large burst of missed frames (more than 20), then use an alternate
                //method to count missed frames.
                if (((VCDUmark0 - VCDUmark1) > 1) && ((VCDUmark0 - VCDUmark1) <= 20)){
                    if (!MissedFrameInitialState){//EV
                        // Don't count frame errors as misses
                        int framemisses = VCDUmark0 - VCDUmark1 - 1 - consecutive_frame_errors;
                        if (framemisses > 0) {
                            GR_LOG_INFO(d_debug_logger,
                                        boost::format("Missed frames: delta VCDU = %d, misses = %d)")
                                          % (VCDUmark0 - VCDUmark1) % framemisses);
                            Missed_FrameCount += framemisses;
                        }
                    }
                    else{
                        MissedFrameInitialState = false;
                    }
                }

                //This alternate method counts how many bytes you processed without finding a frame synch flag
                else if ((bad_data_count > 0) && ((VCDUmark0 - VCDUmark1) > 20)) {
                    if (!MissedFrameInitialState){//EV
                        int framemisses = (int)(bad_data_count / 1020.0) - consecutive_frame_errors;
                        if (framemisses > 0) {
                            GR_LOG_INFO(d_debug_logger,
                                        boost::format("Missed frames: delta VCDU = %d, misses = %d)")
                                          % (VCDUmark0 - VCDUmark1) % framemisses);
                            Missed_FrameCount += framemisses;
                            // printf("Missed %d Bad data Frames\n",int((double)this->bad_data_count/1020.0));
                            //Don't double count frame misses and frame errors;
                        }
                    }
                    else{
                        MissedFrameInitialState = false;
                    }
                }
                consecutive_frame_errors = 0;
                bad_data_count=0;  //Reset this to zero since this is the if(syncsum <= syncthesh) code block.

                GR_LOG_DEBUG(d_debug_logger, boost::format("Received VCID %d") % VCID);
                if (VCID == 63){
                    return 0;
                }

                if (VCIDtable[VCID][8] == 0)
                    VCIDtable[VCID][8] = VCDUcount;
                else
                {
                    if (VCDUcount != VCIDtable[VCID][8] + 1)
                    {// Skipped a packet
                        if (VCIDtable[VCID][2] != 2049)
                            APIDtable[VCIDtable[VCID][2]] = 2040;
                        VCIDtable[VCID][0] = 0;
                        VCIDtable[VCID][1] = 0;
                        VCIDtable[VCID][2] = 2049;  // OUT OF RANGE
                        VCIDtable[VCID][3] = 2047;  // Packet contains data from previous block
                        VCIDtable[VCID][4] = 0;
                        VCIDtable[VCID][5] = 0;
                        VCIDtable[VCID][6] = 0;
                        VCIDtable[VCID][7] = 0;
                    }
                    VCIDtable[VCID][8] = VCDUcount;
                }


                Fullheader = VCIDtable[VCID][0];
                Extension = VCIDtable[VCID][1];
                APIDindex = VCIDtable[VCID][2];
                PendingAPID = VCIDtable[VCID][3];
                sequenceflag = VCIDtable[VCID][4];
                APID = VCIDtable[VCID][7];
                head = false;
                firstheaderpointer = LRITframe[6];
                firstheaderpointer = ((firstheaderpointer << 8) | LRITframe[7]) & 0x7ff;
                framepointer = 8;
                remainingframebytes = 884;
                if (firstheaderpointer == 2047)
                {
                    if (PendingAPID == 2047){
                        VCIDtable[VCID][0] = 0;
                        VCIDtable[VCID][1] = 0;
                        VCIDtable[VCID][2] = 2049;//Set APID index out of range
                        VCIDtable[VCID][3] = 2047;//Set Pending Apid=2047
                        VCIDtable[VCID][4] = 0;
                        VCIDtable[VCID][5] = 0;
                        VCIDtable[VCID][6] = 0;
                        VCIDtable[VCID][7] = 0;
                        return 0;
                    }

                    if ((!Fullheader) && Extension)
                    {
                        GetLritPacketHeader();
                        if (APID == 2047){ // Pending Apid=2047
                            if (PacketLength < remainingframebytes)
                            {
                                framepointer += PacketLength;
                                remainingframebytes -= PacketLength;
                                //goto getapid;
                                getAPID = true;
                            }else if (PacketLength == remainingframebytes) {
                                VCIDtable[VCID][1] = 0;
                                return 0;
                            }
                            if (getAPID==false){
                                VCIDtable[VCID][1] = 1;
                                VCIDtable[VCID][3] = 2047;
                                Bytesremaining[2047] = PacketLength - remainingframebytes;
                                return 0;
                            }
                        }
                        if (getAPID==false){
                            PendingAPID = VCIDtable[VCID][2];
                            head = true;
                        }
                    }

                    if (getAPID==false){
                        if (!Extension)
                            return 0;

                        APIDindex = PendingAPID;
                        if (Bytesremaining[APIDindex] < 884 - (framepointer - 8))
                            return 0;
                        if (head)
                        {
                            if (((TMP_BUFFER_SIZE - (LRITFile[APIDindex] - LRITFilePtr[APIDindex])) < 6) || (APIDindex > NUMBER_OF_FILES)){
                                GR_LOG_WARN(d_logger, boost::format("TMP Buffer Overload or invalid APIDindex=%i ERROR 1") % APIDindex);
                                return 0;
                            }
                            memcpy(LRITFilePtr[APIDindex], &LRIThead[0], 6);
                            LRITFilePtr[APIDindex] += 6;
                        }
                        if (((TMP_BUFFER_SIZE - (LRITFile[APIDindex] - LRITFilePtr[APIDindex])) < (884 - (framepointer - 8))) || (APIDindex > NUMBER_OF_FILES)){
                            GR_LOG_WARN(d_logger, boost::format("TMP Buffer Overload or invalid APIDindex=%i ERROR 2") % APIDindex);
                            return 0;
                        }
                        memcpy(LRITFilePtr[APIDindex], &LRITframe[framepointer], 884 - (framepointer - 8));
                        LRITFilePtr[APIDindex] += 884 - (framepointer - 8);
                        Bytesremaining[APIDindex] -= (884 - (framepointer - 8));
                        return 0;
                    }
                }

                if ((firstheaderpointer > 0)||(!getAPID)){
                    if ((!Fullheader) && Extension)
                    {
                        GetLritPacketHeader();
                        PendingAPID = VCIDtable[VCID][2];
                        head = true;
                    }

                    if (Extension) {
                        APIDindex = PendingAPID;
                        if (head) {
                            if (((TMP_BUFFER_SIZE - (LRITFile[APIDindex] - LRITFilePtr[APIDindex])) < 6)||(APIDindex > NUMBER_OF_FILES)){
                                GR_LOG_WARN(d_logger, boost::format("TMP Buffer Overload or invalid APIDindex=%i ERROR 3") % APIDindex);
                                return 0;
                            }
                            memcpy(LRITFilePtr[APIDindex], &LRIThead[0], 6);
                            LRITFilePtr[APIDindex] += 6;
                        }

                        if (((TMP_BUFFER_SIZE - (LRITFile[APIDindex] - LRITFilePtr[APIDindex])) < Bytesremaining[APIDindex]) || (APIDindex > NUMBER_OF_FILES)){
                            GR_LOG_WARN(d_logger, boost::format("TMP Buffer Overload or invalid APIDindex=%i ERROR 4") % APIDindex);
                            return 0;
                        }

                        memcpy(LRITFilePtr[APIDindex], &LRITframe[framepointer], Bytesremaining[APIDindex]);
                        LRITFilePtr[APIDindex] += Bytesremaining[APIDindex];

                        if ((sequenceflag == 3) || (sequenceflag == 2)){
                            ProcessLRITfile(LRITFilePtr[APIDindex] - LRITFile[APIDindex]);
                            LRITFilePtr[APIDindex] = LRITFile[APIDindex];
                            APIDtable[APIDindex] = 2040;
                            filesequence[APID]++;
                        }
                        remainingframebytes -= Bytesremaining[APIDindex];
                        framepointer += Bytesremaining[APIDindex];
                        //goto getapid;
                    } else{
                        framepointer += firstheaderpointer;
                        remainingframebytes -= firstheaderpointer;
                        //goto getapid;
                    }
                    getAPID = true; // It's enable for both cases of (Extension)
                }

                //if (firstheaderpointer == 0)
                if ((firstheaderpointer == 0) ||(getAPID) ){ //will have a complete packet header available
                    //getapid:
                    getAPID = true;

                    while ( getAPID == true){
                        //0-Fullheader,1-extension,2-APIDindex,3-pendingapid,4-sequence,5-packetindex,6-packetlength
                        if (!GetLritPacketHeader())
                            return 0;

                        if (APID == 2047){//2047=Fill APID - ignore
                            if (PacketLength < remainingframebytes) {
                                framepointer += PacketLength;
                                remainingframebytes -= PacketLength;
                                getAPID = true;
                                //goto getapid;
                            }else{
                                //getAPID = false;
                                if (PacketLength == remainingframebytes){
                                    VCIDtable[VCID][1] = 0; //Extension=false;
                                    return 0;
                                }
                                VCIDtable[VCID][1]  = 1;//Extension=true;
                                VCIDtable[VCID][3]  = 2047;//PendingAPID=2047;
                                Bytesremaining[2047] = PacketLength - remainingframebytes;
                                return 0;
                            }
                        }else{
                            getAPID = false;
                        }

                        if (getAPID == false){
                            ////Add Bytes to File///
                            if (Bytesremaining[APIDindex] <= remainingframebytes) {
                                bytestowrite = Bytesremaining[APIDindex];
                                VCIDtable[VCID][1] = 0;
                            } else{
                                bytestowrite = remainingframebytes;
                                Bytesremaining[APIDindex] -= remainingframebytes;
                                PendingAPID = VCIDtable[VCID][2];
                                VCIDtable[VCID][3] = PendingAPID;
                                VCIDtable[VCID][1] = 1;
                            }
                            if (((TMP_BUFFER_SIZE - (LRITFile[APIDindex] - LRITFilePtr[APIDindex])) < bytestowrite + 6) || (APIDindex > NUMBER_OF_FILES)){
                                GR_LOG_WARN(d_logger, boost::format("TMP Buffer Overload or invalid APIDindex=%i ERROR 5") % APIDindex);
                                return 0;
                            }
                            //EV This is where LRIT usually crashed
                            //EV Add
                            if (APIDindex<NUMBER_OF_FILES){ //this doesn't fix it
                                memcpy(LRITFilePtr[APIDindex], &LRITframe[framepointer - 6] ,bytestowrite + 6);
                            }else{
                                GR_LOG_WARN(d_logger, boost::format("APID Index= %i exceeds array size (%i), EXIT") % NUMBER_OF_FILES % APIDindex);
                                return 0;
                            }
                            LRITFilePtr[APIDindex] += bytestowrite + 6;
                            remainingframebytes -= bytestowrite;
                            framepointer += bytestowrite;
                            if (VCIDtable[VCID][1])
                                return 0;
                            if ((sequenceflag == 3) || (sequenceflag == 2)){
                                ProcessLRITfile(LRITFilePtr[APIDindex] - LRITFile[APIDindex]);
                                APIDtable[APIDindex] = 2040;
                                filesequence[APID]++;
                                LRITFilePtr[APIDindex] = LRITFile[APIDindex];
                            }
                            if (remainingframebytes <= 0){
                                return 0;
                            }
                            //goto getapid;
                            getAPID = true;
                        }
                    }
                }
            }

            //At the end an iteration of the bit offset loop
            //If this portion of code is reached, that means that no header was found during
            //this particular bit marker


            //Estimate and print what the missed synch header was.  Its not perfect, so disregard outliers reported
            if ((syncsum > syncthresh) && (offset==best_offset) && (best_offset_ready_signal==1))
            {
              // printf("Synched ? %d\n",NoSyncFound==0);
              GR_LOG_INFO(d_debug_logger,
                          boost::format("Header Errors Exceeded Threshold: "
                                        "Sync Sum / thresh = %d / %d, Bit Offset = %d")
                                        % syncsum % syncthresh % best_offset);
              //subtract_four = true;
              best_offset_ready_signal=0;
            }

            //Update, 32 bit sequence to search Continue searching for frame sync...bit by bit
            Ltreg = Ltreg << 1;
            if (hreg & 0x80)    // hreg has the next input byte
                Ltreg++;
            hreg = hreg << 1;
        }

       //Checked every bit offset within for a given byte marker without finding a header

        LastProcBytePtr++; //No sync...get next byte, and search every offset for the header
        InvByteCount++;  //Update the total number of bytes processed without finding a header
        consecutive_good_frames = 0;


        //If you are near the next frame header, relax the threshold
        if (((InvByteCount >= (1024-SEARCH_DEPTH)) && (InvByteCount <= (1024+SEARCH_DEPTH))) && ((consecutive_frames_missed+consecutive_frame_errors) < 2)){
            if (lock_status == 1){
                GR_LOG_INFO(d_debug_logger,
                             boost::format("Lock Status Changed from %d to %d")
                                            % lock_status % 2);
                lock_status = 2;
            }
        }
        //If you are searching for a header in the middle of a packet, downgrade the lock
        //status
        else if (lock_status>1){
            GR_LOG_INFO(d_debug_logger,
                         boost::format("Lock Status Changed from %d to %d")
                                       % lock_status % 1);
            lock_status = 1;
        }

        BadData=false;
        if (InvByteCount > 1024){ //Checked all 1024 bytes, still no header found
            if (InvToggleFlag){
                GR_LOG_INFO(d_debug_logger, "No header in 1K bytes (InvToggleFlag): toggle data inversion");
                LRITDataInvert = LRITDataInvert ^ 0xff;
                InvToggleFlag = false;
                InvByteCount = 0;
                invresetpointer = LastProcBytePtr;
                InvBadFrameCount = 0;
                BadData = true;

            }
            if ((foundsyncflag)&&(!BadData)) {
                InvBadFrameCount++;
                if (InvBadFrameCount >= 3)
                {
                    GR_LOG_INFO(d_debug_logger, "Bad frames: toggle data inversion");
                    LastProcBytePtr = invresetpointer;
                    LRITDataInvert = LRITDataInvert ^ 0xff;
                    InvToggleFlag = true;
                    InvBadFrameCount = 0;
                    foundsyncflag = false;
                }
                BadData = true;
            }
            if (!BadData){
                GR_LOG_INFO(d_debug_logger, "No header in 1K bytes (BadData): toggle data inversion");
                LastProcBytePtr = invresetpointer;
                LRITDataInvert = LRITDataInvert ^ 0xff;
                InvToggleFlag = true;
                InvByteCount = 0;
            }
        }
        BadData=true;

        bad_data_count++;
#ifdef V2_2
        num_bad++;
#endif
        consecutive_frames_missed = consecutive_frames_missed + int(bad_data_count/1020.0);
    }

    return 0;
}

int CGoesProcessor::EMWIN_Main(int resynch_thresh)
{
    unsigned char hreg;
    char tempstr1[200];
    int offset;
    int EndofEmwinDataPtr;
    bool head = false;
    int bytestowrite = 0;
    char xcor;
    unsigned int i, j, k, m, n, ercount, ercountmax = 0;
    int LRcount;
    unsigned long synk;
    int syncsum,syncthresh;
    unsigned short temword;
    FILE *BER_File;
    int SEARCH_DEPTH = 5;  //how far you can search for a header in bytes, before downgrading lock status
    int CONSECUTIVE_GOOD_THRESHOLD = 3; //How many consecutive frames I need to receiver in order to be in the highest lock state.
    int CONSECUTIVE_BAD_THRESHOLD = 3;
    int resynch_thresh3 = resynch_thresh;  //Found more than CONSECUTIVE THRESHOLD Frames in the Exact positions you expect to find headers
    int resynch_thresh2 = 6; //Had one missed frame or frame in error but looking in areas where you expect to find headers.
    int resynch_thresh1 = 2; //Looking for header in the middle of a packet
    if (resynch_thresh2 > resynch_thresh3)
        resynch_thresh2 = resynch_thresh3;

    //=======================================
    // Eliminate GOTOs
    bool BadData= false;
    //=======================================
    //                   EMWIN-OQPSK Coded VCDU
    //
    // ASM          VCDU         BPDU           BPDU Data Field   Reed-Solomon
    //0x1ACFFC1D        Header       Header                               Check Symbols
    // 4 octets    6 octets      2 octets       884 octets         128 octets
    //<<<----------------------------  1024 bytes      ---------------------->>

#ifdef V2_2
    if (num_bad > delay_thresh) {
      if (use_delay) { toggle_viterbi_delay(); }
      else { toggle_rotate(); }
      num_bad = 0;
    }
#endif


    if ((CurrentDataPtr - LastProcBytePtr) >= MIN_DATA_SIZE)
    {
        //--------------------
        //      ASM HEADER
        //--------------------
        if (LRITFirstsync)
        {   //   ^ = OR
            // LRITDataInvert ; Alternates between all 0's, all 1's
            // Ltreg = FFFF FFFFxh
            for (i = 0; i < 4; i++)
                Ltreg = (Ltreg << 8) | ((*(LastProcBytePtr + i)) ^ LRITDataInvert);
            LastProcBytePtr += 4;
            LRITFirstsync = false;
        }

        hreg = *LastProcBytePtr ^ LRITDataInvert;
        for (offset = 0; offset <= 7; offset++)
        {
            synk = lritsync ^ Ltreg;
            //-----------------------------------------------------
            // Bitwise comparison of lritsync  and Ltreg= rec.data
            // if they are both equal, then the XOR operation results in
            // synk ==0, which later makes suncsum =0, which implies that the
            // header was found

            syncsum = 0;
            for (n = 0; n < 32; n++)
            {
                if (synk & 0x00000001)
                    syncsum++;
                synk = synk >> 1;
            }

            if ((consecutive_frames_missed + consecutive_frame_errors) > CONSECUTIVE_BAD_THRESHOLD){
                if (lock_status != -1)
                    GR_LOG_INFO(d_debug_logger, boost::format("Lock Status Changed %d --> %d") % lock_status % -1);
                lock_status = -1;
            }

            if (lock_status==1)
                syncthresh = resynch_thresh1;
            else if (lock_status==2)
                syncthresh = resynch_thresh2;
            else if (lock_status>=3 && (offset!=best_offset)) //Wrong expected bit alignment within bytesteam
                syncthresh = resynch_thresh2;
            else if (lock_status>=3 && (offset==best_offset)) //Correct expected bit alignment within bytesteam
                syncthresh = resynch_thresh3;
            else if (lock_status<0)
                syncthresh = 0;

            if (syncsum <= syncthresh){
                //--------------------
                //  FOUND  ASM HEADER
                FrameSyncCnt ++;
                best_offset_ready_signal = 1; //This is to allows you to estimate and print the syncsum for the undetected headers.
                if (consecutive_good_frames >= CONSECUTIVE_GOOD_THRESHOLD){
                    if (lock_status != 3)
                        //printf("Lock Status Changed From %d to %d\n", lock_status, 3);
                        lock_status = 3;  //upgrade lock status directly to 3
                    best_offset = offset;
                }


                InvByteCount = 0;
                InvBadFrameCount = 0;
                invresetpointer = LastProcBytePtr;
                foundsyncflag = true;
                LRITFirstsync = true;
                if (!LRITsyncflag)
                    LRITsyncflag = true;

                if (!Firstsyncflag)
                {
                    processbytes = (int)(LastProcBytePtr - old_LastProcBytePtr);
                    if (processbytes > 1024)
                        LRITsyncflag = false;
                }
                else
                    Firstsyncflag = false;

                old_LastProcBytePtr = LastProcBytePtr;

                //Randomizer: Frame size
                for (j = 0; j < LRIT_PACKET_SIZE; j++)
                {
                    LRITframe[j] = (*(LastProcBytePtr + j)) ^ LRITDataInvert;
                    if (offset != 0)
                    {
                        temword = (*(LastProcBytePtr + j)) ^ LRITDataInvert;
                        temword = (temword << 8) | (*(LastProcBytePtr + j + 1)) ^ LRITDataInvert;
                        LRITframe[j] = temword >> (8 - offset);
                    }
                    LRITframe[j] ^= Random[j];
                }

                LastProcBytePtr += LRIT_PACKET_SIZE;
                EMWINFrameCount++;

                for (k = 0; k < 1024; k++)
                    LRIT1frame[k] = LRITframe[k];

                //RS Decode
                LRITrs->Decode((unsigned char *) &LRITframe[0]);
                viterbi_framerrors = LRITrs->framerrors;
                RS_framerrors=-100;
                RS_framerrors = LRITrs->uncorrected_framerrors;
                ercount=0;
                for (k = 0; k < 1020; k++)
                {
                    xcor = LRIT1frame[k] ^ LRITframe[k];
                    for (m = 0; m < 8; m++)
                    {
                        if ((xcor & 0x01) == 0x01)
                        {
                            ercount++;
                        }
                        xcor = xcor >> 1;
                    }
                }
                ErCountCum = ErCountCum + ercount;

                //If there is a Reed Solomon error, update error counters
                //Change Frame Synchronizatio Status
                // Try and recover the VCDU counter so you can detect a missed frame
                if (RS_framerrors>0)
                {
                    EMWINFrameError+= RS_framerrors;
                    consecutive_frame_errors++;
                    consecutive_good_frames = 0;
                    VCID = LRITframe[1] & 0x3f;
                    VCDUcount = LRITframe[2];
                    VCDUcount = (VCDUcount << 8) | LRITframe[3];
                    VCDUcount = (VCDUcount << 8) | LRITframe[4];
                    //printf(" \n Frame %d has Reed Solomon Error. Est. VCDU=%d Syncsum%d\n" ,LRITFrameCount,VCDUcount,syncsum);
                    if (lock_status==3){ //Strictly downgrade from 3 to 2 only.
                        //printf("Lock Status Changed From %d to %d\n", 3, 2);
                        lock_status=2;
                    }
                    return 0;
                }
                //The rest of code in [if(syncsun<sync_thresh)] represents a frame synch and a good received frame

                consecutive_frames_missed = 0;
                if (lock_status<2){ //If found a and synched to good from, upgrade from 0 or 1, up to 2.
                    //printf("Lock Status Changed From %d to %d\n", lock_status, 2);
                    lock_status = 2;
                }

                consecutive_good_frames++;

                //Table 2 2 VCDU Header Structure
                //Version               VCDU-ID        VCDU          Signaling Field
                //Number            SCID        VCID     Counter        Replay Flag       Spare
                //2 bits      8 bits   6 bits    24 bits    1 bit         7 bits

                // 1st Byte = 2 bits from Version Number, 6 from SCID
                // 2nd Byte = 2 bits from SCID, all 6 bits from VCID
                VCID = LRITframe[1] & 0x3f;     // Grab 6 bits of 8
                // 3rd byte = VCDU Counter
                VCDUcount = LRITframe[2];
                VCDUcount = (VCDUcount << 8) | LRITframe[3];
                VCDUcount = (VCDUcount << 8) | LRITframe[4];
                VCDUmark1= VCDUmark0;
                VCDUmark0 = VCDUcount;

                if ((EMWINFrameCount % 20) == 0 || syncsum>=8) {
                  GR_LOG_INFO(d_logger,
                               boost::format("Total FER=%-8.3eRS FER=%-8.3e"
                                             "Vit.BER=%-8.3eViterbi Errs=%-8d"
                                             "RS Frame Errors=%-8ldMissed VCDU Frames=%-8ld"
                                             "Total VCDU Frames=%-8ldEMWIN")
                                    % ((EMWINFrameError+((double)Missed_FrameCount)*4.0) /
                                        ((double)EMWINFrameCount*4.0))
                                    % (EMWINFrameError / ((double)EMWINFrameCount*4.0))
                                    % ((double)ErCountCum / (double)(EMWINFrameCount*8192))
                                    % ErCountCum % EMWINFrameError
                                    % Missed_FrameCount % EMWINFrameCount);
                }

                //IF the VCDU markers do not agree, declare a missed frame.
                //If there is a large burst of missed frames (more than 20), then use an alternate
                //method to count missed frames.
                if (((VCDUmark0 - VCDUmark1) > 1) && ((VCDUmark0 - VCDUmark1) <= 20)){
                    if (!MissedFrameInitialState){//EV
                        if (((VCDUmark0 - VCDUmark1)-1 - consecutive_frame_errors) > 0){
                            Missed_FrameCount = Missed_FrameCount + (VCDUmark0 - VCDUmark1)-1 - consecutive_frame_errors;
                            //printf("Missed %d VCDU Frames\t VCDUmark0 = %d\t VCDUmark1 = %d\n",(VCDUmark0 - VCDUmark1)-1, VCDUmark0,VCDUmark1);
                            //Don't double count frame misses and frame errors
                        }
                    }
                    else{
                        MissedFrameInitialState = false;
                    }
                }

               //This alternate method counts how many bytes you processed without finding a frame synch flag
                else if ((this->bad_data_count >0) && ((VCDUmark0 - VCDUmark1) > 20)){
                    if (!MissedFrameInitialState){//EV
                        if ((int((double)this->bad_data_count/1020.0)- consecutive_frame_errors)>0){
                            Missed_FrameCount = Missed_FrameCount + int((double)this->bad_data_count/1020.0)- consecutive_frame_errors;
                            // printf("Missed %d Bad data Frames\n",int((double)this->bad_data_count/1020.0));
                            //Don't double count frame misses and frame errors;
                        }
                    }
                    else{
                        MissedFrameInitialState = false;
                    }
                }

                consecutive_frame_errors = 0;
                this->bad_data_count=0;  //Reset this to zero since this is the if(syncsum <= syncthesh) code block.


                // VCID Binary 001111 specifying channel 15

                if (VCID != 15)
                    return 0;

                // BPDU Data Field = 884 octets = 7071 bits
                EndofEmwinDataPtr = LRITframe[6] & 0x3f;
                //BUG:: The spec eludes to using 0x07, but this causes errors!  0x3f works better for some reason?!

                EndofEmwinDataPtr = (EndofEmwinDataPtr << 8) | LRITframe[7];//JR [Grab remaining 8 bits (in the 11 bit Header Pointer field)
                if ((EndofEmwinDataPtr == 16383) || (EndofEmwinDataPtr == 7071)) {

                    if ((EmwinBufferLeft = (EmwinBufEnd - EmwinPtr)) >= 884)
                    {
                        memcpy(EmwinPtr, &LRITframe[8], 884);
                        EmwinPtr += 884;
                        EmwinBytes += 884;
                    }
                    else  //emwindata[10000000] full, start over
                    {
                        memcpy(EmwinPtr, &LRITframe[8], EmwinBufferLeft);
                        EmwinPtr = EmwinData;
                        EmwinBytes += EmwinBufferLeft;
                        memcpy(EmwinPtr, (&LRITframe[8]) + EmwinBufferLeft, (884 - EmwinBufferLeft));
                        EmwinBufferLeft = (884 - EmwinBufferLeft);
                        EmwinPtr += (EmwinBufferLeft);
                        EmwinBytes += EmwinBufferLeft;
                    }
                }
                else
                {
                    if ((EmwinBufferLeft = (EmwinBufEnd - EmwinPtr)) >= EndofEmwinDataPtr)
                    {
                        memcpy(EmwinPtr, &LRITframe[8], EndofEmwinDataPtr);
                        EmwinPtr += EndofEmwinDataPtr;
                        EmwinBytes += EndofEmwinDataPtr;
                    }
                    else
                    {
                        memcpy(EmwinPtr, &LRITframe[8], EmwinBufferLeft);
                        EmwinPtr = EmwinData;
                        EmwinBytes += EmwinBufferLeft;
                        memcpy(EmwinPtr, (&LRITframe[8]) + EmwinBufferLeft, (EndofEmwinDataPtr - EmwinBufferLeft));
                        EmwinBufferLeft = (EndofEmwinDataPtr - EmwinBufferLeft);
                        EmwinPtr += (EmwinBufferLeft);
                        EmwinBytes += EmwinBufferLeft;
                    }
                }

                ProcessEMWINfile();
                return 0;
            }


            //At the end an iteration of the bit offset loop
            //If this portion of code is reached, that means that no header was found during
            //this particular bit marker.

            //Update, 32 bit sequence to search Continue searching for frame sync...bit by bit
            Ltreg = Ltreg << 1;
            if (hreg & 0x80)
                Ltreg++;
            hreg = hreg << 1;
        }

        //Checked every bit offset within for a given byte marker without finding a header


        LastProcBytePtr++;
        InvByteCount++;
        consecutive_good_frames = 0; //Could not find header in proper location, check
        //If you are near the next frame header, relax the threshold
        if (((InvByteCount >= (1024-SEARCH_DEPTH)) && (InvByteCount <= (1024+SEARCH_DEPTH))) && ((consecutive_frames_missed+consecutive_frame_errors) < 2)){
            if (lock_status == 1){
                lock_status = 2;
            }
        }
        else if (lock_status>1){
            lock_status = 1;
        }

        BadData= false;
        if (InvByteCount > 1024){
            if (InvToggleFlag){
                LRITDataInvert = LRITDataInvert ^ 0xff;
                InvToggleFlag = false;
                InvByteCount = 0;
                invresetpointer = LastProcBytePtr;
                InvBadFrameCount = 0;
                //goto baddata;
                BadData= true;
            }
            if ((foundsyncflag)&&(!BadData)) {
                InvBadFrameCount++;
                if (InvBadFrameCount >= 3){
                    LastProcBytePtr = invresetpointer;
                    LRITDataInvert = LRITDataInvert ^ 0xff;
                    InvToggleFlag = true;
                    InvBadFrameCount = 0;
                    foundsyncflag = false;
                }
                //goto baddata;
                BadData= true;
            }

            if (!BadData){
                LastProcBytePtr = invresetpointer;
                LRITDataInvert = LRITDataInvert ^ 0xff;
                InvToggleFlag = true;
                InvByteCount = 0;
            }
        }

// baddata:
        BadData= true;
        InvToggleFlag = InvToggleFlag;
        this->bad_data_count++;
#ifdef V2_2
        num_bad++;
#endif
        consecutive_frames_missed = consecutive_frames_missed + int((double)this->bad_data_count/1020.0);
    }

    return 0;
}

bool CGoesProcessor::GetLritPacketHeader()
{
    int i;
    bool Done_SaveAPID=false;
    // Grab data from CP_PDU User Data
    for (PacketIndex = VCIDtable[VCID][5]; PacketIndex < 6; PacketIndex++)
    {
        packetheader[VCID][PacketIndex]=LRITframe[framepointer];
        framepointer+=1;
        remainingframebytes-=1;
        if (remainingframebytes <= 0)
        {
            VCIDtable[VCID][5]=PacketIndex+1;
            VCIDtable[VCID][0]=0;
            VCIDtable[VCID][1]=1;
            return false;
        }
    }

    for (i = 0; i < 6; i++)
        LRIThead[i] = packetheader[VCID][i];
    VCIDtable[VCID][5] = 0;

    //EV: this is not really the APID
    // 1st byte has: Version No(3bits), Type(1bit),Secondary Header(1bit), 3 bits from APID
    // 2nd byte has: remaining 8 bits from APID. APID has a total of 11 bits
    APID = packetheader[VCID][0];
    if (APID & 0x08)
        SrchHeaderFlag = true;
    else
        SrchHeaderFlag = false;
    //EV: This is the "real" APID
    APID = ((APID << 8) | packetheader[VCID][1]) & 0x7ff;

    //SequenceFlag = 2bits from 3rd byte
    sequenceflag = packetheader[VCID][2];
    sequenceflag = (sequenceflag >> 6) & 0x03;
    VCIDtable[VCID][4] = sequenceflag;
    VCIDtable[VCID][7] = APID;

    // Packet Sequence Counter
    Seqcount = (((packetheader[VCID][2]) & 0x3f) << 8) | packetheader[VCID][3];

    // Packet Length
    PacketLength = packetheader[VCID][4];
    PacketLength = (PacketLength << 8) | packetheader[VCID][5];
    PacketLength++;   //EV Why this addition ?
    VCIDtable[VCID][6] = PacketLength;
    VCIDtable[VCID][0] = 1;

    if (APID == 2047)
        return true;

    //Checks if APID of current packet is in the APIDtable, and if so sets bytesremaining for that APID to the packet length,
    // stores the APID index in VCIDtable, then returns
    for (APIDindex = 0; APIDindex < MAX_APID_TABLE; APIDindex++)
    {
        if (APID == APIDtable[APIDindex])
        {
            Bytesremaining[APIDindex] = PacketLength;
            VCIDtable[VCID][2] = APIDindex;
            return true;
        }
    }

    //If continuation or end packet, then there's no header
    if ((sequenceflag != 3) && (sequenceflag != 1))
        return false;

    Done_SaveAPID=false;

    //Got to this point if the APID for this packet hasn't yet been encountered
    //Loop below finds the first unused entry in the APIDtable, and stores the APID of the current packet in that entry.
    //Also updates the bytesremaining and stores the APIDindex in the VCID Table
    for (APIDindex = 0; APIDindex < MAX_APID_TABLE; APIDindex++) {
        if (APIDtable[APIDindex]==2040) {
            APIDtable[APIDindex]=(unsigned short)APID;
            Bytesremaining[APIDindex]=PacketLength;
            VCIDtable[VCID][2]=APIDindex;
            if (APIDindex>NUMBER_OF_FILES) return false;  //Why was this if(APIDindex>19)????????
            LRITFilePtr[APIDindex]=LRITFile[APIDindex];
            //goto donesaveAPID;
            Done_SaveAPID=true;
        }
        if (Done_SaveAPID){
            break;
        }
    }
//donesaveAPID:
    Done_SaveAPID=true;
    return true;
}

void CGoesProcessor::ProcessLRITfile(int bytes)
{
    char tempstr[200], tempstr1[20], tempstr2[20];
    unsigned long dwBytesRead;
    int Error;
    LRIThavename = false;
    sprintf(tempstr1, "%d", APID);
    sprintf(tempstr2, "%d", filesequence[APID]);
    sprintf(tempstr, "LRITtemp/LRIT%s_%s", tempstr1, tempstr2);

    lpBufPtr = LRITFile[APIDindex];

    if (CalcCRC(bytes))
    {
        GR_LOG_INFO(d_debug_logger,
                    boost::format("LRIT File: APID # = %i, type = %i, name = %s") %
                    APIDindex % LRITfiletype % LRITfilename);
        // Some of them are type DCS ???
        if ((LRITfiletype != 0) && (LRITfiletype != 130)){
            //Intermediate files ??
            //TYPE 214: Intermediate LRIT
            //TYPE 130: DCS File
            return;
        }

        // create a scope oustide of which file is closed automagically
        {
            std::ostringstream makefilepath;
            makefilepath << lritdir << '/';
            if (LRIThavename){
                //OR_ABI... named segments all have the exact same filename, need something to
                //differentiate them so they don't get overwritten
                std::string fnstr = std::string(LRITfilename);
                if(fnstr.substr(fnstr.find_last_of(".") + 1) == "lrit" && LRITfiletype == 0) {
                    makefilepath << fnstr.substr(0,fnstr.length()-5)<<"_Seg" << SegNum << ".lrit";
                }
                else{
                    makefilepath << LRITfilename;
                }

            }
            else{
              makefilepath << "LRIT" << APID << "_" << filesequence[APID];
            }

            std::string filepath(makefilepath.str());
            std::ofstream outfile(filepath, std::ios::binary);
            if (!outfile) {
                GR_LOG_ERROR(d_logger, boost::format("Could not write output file %s") % filepath);
            }
            else {
                outfile.write(reinterpret_cast<char*>(FileBuffer), numLRITbytes);
                GR_LOG_INFO(d_logger, "Extracted " + filepath);
            }
        }

        haveLRITfiles = true;
    }
 return;
}

void CGoesProcessor::ProcessEMWINfile()
{
    unsigned char *Headerptr;
    int seq, MaxBlks, index;
    char filename[FILENAME_SIZE+1];
    char headerbuff[HEADER_SIZE];
    char temstr[20];

    //======================================
    // Eliminate GOTOs
    bool ContinueStartSearch=false;
    bool DoneBlock = false;
    //======================================
    //EmwinStart is initialized as false
    if (!EmwinStart){
        while ( (EmwinProcBytes < EmwinBytes)&&(!EmwinStart)){
            if (0 != *EmwinProcPtr){
                EmwinProcPtr++;
                if (EmwinProcPtr > EmwinBufEnd)
                    EmwinProcPtr = EmwinData;
                EmwinProcBytes++;
                EmwinNullCount = 0;
                //continue may only be used inside a loop, such as a for statement. It terminates the current iteration of the loop and proceeds directly to the next. In the case of a for loop it jumps to its increment-expr.
                continue;
            }

            EmwinNullCount++;
            EmwinProcPtr++;
            if (EmwinProcPtr > EmwinBufEnd){
                EmwinProcPtr = EmwinData;
            }
            EmwinProcBytes++;
            if (EmwinProcBytes >= EmwinBytes){
                return;
            }

            ContinueStartSearch = false;
            while ((EmwinNullCount < 6)&&(!ContinueStartSearch)){//6 null characters point to start of packet
                if (0 == *EmwinProcPtr){ //Found another one. Increment count
                    EmwinNullCount++;
                    EmwinProcPtr++;
                    if (EmwinProcPtr > EmwinBufEnd){
                        //Wrap back to beginning of buffer
                        EmwinProcPtr = EmwinData;
                    }
                    EmwinProcBytes++;
                    if (EmwinProcBytes >= EmwinBytes){
                        //need more data
                        return;
                    }
                    ContinueStartSearch = false;
                }else{
                    // Found a null by chance, but not 6 in a row.
                    // Restart Search procedure
                    EmwinNullCount = 0;
                    EmwinProcPtr++;
                    if (EmwinProcPtr > EmwinBufEnd){
                        //Wrap back to beginning of buffer
                        EmwinProcPtr = EmwinData;
                    }
                    EmwinProcBytes++;
                    //goto continuestartsearch;
                    ContinueStartSearch = true;
                }
            }

            if (!ContinueStartSearch){
                while ('/' != *EmwinProcPtr){
                    EmwinProcPtr++;
                    if (EmwinProcPtr > EmwinBufEnd)
                        EmwinProcPtr = EmwinData;
                    EmwinProcBytes++;
                    if (EmwinProcBytes >= EmwinBytes)
                        return;
                }
                EmwinStart = true;
                EmwinNullCount = 0;
                EmwinPacketCount = 0;
                //goto foundstart;
                //FoundStart=true;  // ---> This is actuall the same as EmwinStart ???
            }

            //continuestartsearch:
            ContinueStartSearch=true;
            //Not used  i = 0;
        }
    }

    //foundstart:
    // EMWIN Packet: 1116 bytes
    //   6 Header =(NULL ) bytes

    if ((1110) > (EmwinBytes - EmwinProcBytes)) //Save 1110
        return;

    static int PrevRemain = 0;
    PrevRemain = remain;

    if ((remain = EmwinBufEnd - EmwinProcPtr) >= (1110+MIN_DATA_SIZE))  //Ensure that remain > 1110
    {// Get FILENAME
        // PF+ 12 bytes for filename
        // HEADER_SIZE bytes worth of header fields, sections 2 - 8 in EMWIN spec - winpro.htm
        memcpy(headerbuff, EmwinProcPtr, HEADER_SIZE);
        if (('P' != *(headerbuff+1)) || ('F' != *(headerbuff + 2))){
            //goto doneblock;
            DoneBlock=true;
        }else{
            EmwinProcPtr += HEADER_SIZE;
            EmwinProcBytes += HEADER_SIZE;
        }

    }else{
        memcpy(headerbuff, EmwinProcPtr, HEADER_SIZE);
        memcpy(EmwinData, EmwinProcPtr+HEADER_SIZE, remain-HEADER_SIZE);
        EmwinProcPtr = EmwinData+remain-HEADER_SIZE;
        if (('P' != *(headerbuff+1)) || ('F' != *(headerbuff + 2))){
            DoneBlock=true;
            //goto doneblock;
        }else{
            EmwinProcBytes += HEADER_SIZE;
        }
    }
    if (!DoneBlock){
      Headerptr = (unsigned char*)headerbuff;
      memcpy(filename, Headerptr + 3, FILENAME_SIZE);
      Headerptr += 18;
      index = strcspn((const char*)Headerptr, "/");
      if (index>19){
        GR_LOG_WARN(d_logger, "EMWIN Header is Corrupted");
            index=19;
            return;
        }
        memcpy(temstr, Headerptr, index);
        temstr[index] = 0;
        seq = atoi(temstr);
        Headerptr += index + 3;

        index = strcspn((const char*)Headerptr, "/");
        if (index>19){
            GR_LOG_WARN(d_logger, "EMWIN Header is Corrupted");
            index=19;
            return;
        }
        memcpy(temstr, Headerptr, index);
        temstr[index]   = 0;
        MaxBlks         = atoi(temstr);   // This tells how many parts the file has

        if (seq == 1) { // File has only one part --> usually txt and small ZIS files
            for (int i = 0; i < NUMBER_OF_FILES; i++)
            {// See which FileIndex position is available to store the file
                if (EmwinFileIndex[i][0] == 0){//found available slot
                    EmwinFileIndex[i][0] = 1;//mark in use
                    EmwinFileIndex[i][1] = 1;//block counter=1, 1st part
                    EmwinFileIndex[i][2] = MaxBlks;//total blocks in file
                    EmwinFileIndex[i][3] = 0;//reset use counter

                    memcpy(EmwinFileName[i], filename, FILENAME_SIZE);
                    EmwinPtrTable[0][i] = EmwinProcPtr;   //save pointer to data block
                    // BUG: Should this be [i][0] ?

                    if (seq == MaxBlks){// only 1 block.  Go process
                        EmwinDisplay(i);
                        EmwinFileIndex[i][0] = 0;  // Clean the File Index Array
                        EmwinFileIndex[i][3] = 0; //reset use counter
                        //goto doneblock;  //Not needed
                        //DoneBlock=true;
                    }
                    DoneBlock=true;
                    break;
                    //goto doneblock;
                }
                // If you are here, it's because  EmwinFileIndex[i][0] != 0
                // The paramater WAIT_TIL_DISCARD says that if "WAIT_TIL_DISCARD" pieces
                // have gone by and still this file is inuse, then it's probably lost
                // so discard it !
                if (!DoneBlock){
                    EmwinFileIndex[i][3]++;
                    if (EmwinFileIndex[i][3] > WAIT_TIL_DISCARD){
                        memcpy(temstr, EmwinFileName[i], FILENAME_SIZE);
                        GR_LOG_WARN(d_logger, boost::format("TIMEOUT: File %s has missing portions: Discard File") % temstr);
                        EmwinFileIndex[i][0] = 1;
                        EmwinFileIndex[i][1] = 1;//block counter=1, 1st part
                        EmwinFileIndex[i][2] = MaxBlks;
                        EmwinFileIndex[i][3] = 0;
                        memcpy(EmwinFileName[i], filename, FILENAME_SIZE);
                        EmwinPtrTable[0][i] = EmwinProcPtr;
                        if (seq == MaxBlks){
                            EmwinDisplay(i);
                            EmwinFileIndex[i][0] = 0;
                            EmwinFileIndex[i][3] = 0;
                            //goto doneblock; Not needed
                        }
                        //goto doneblock;
                        DoneBlock=true;
                        break;
                    }
                }
            }
        }else{
            for (int i = 0; i < NUMBER_OF_FILES; i++)
            {// Check to see if it's a pre-existing file
                if (0 == memcmp(filename, EmwinFileName[i], FILENAME_SIZE)) {
                    if (0 == EmwinFileIndex[i][0]){
                        //goto doneblock;//invalid file block
                        DoneBlock=true;
                        break;
                    }

                    if ((seq == EmwinFileIndex[i][2])&&(!DoneBlock)){ //last block:  go process
                        EmwinPtrTable[seq - 1][i] = EmwinProcPtr; //save pointer to data block
                        EmwinDisplay(i);
                        EmwinFileIndex[i][0] = 0;//mark available
                        EmwinFileIndex[i][3] = 0;//reset use counter
                        //goto doneblock;
                        DoneBlock=true;
                        break;
                    }

                    if (  ((seq != (1 + EmwinFileIndex[i][1])) || (seq > 200)) && (!DoneBlock)){
                        //missed some data or file too large.  abort
                        memcpy(temstr, EmwinFileName[i], FILENAME_SIZE);
                        GR_LOG_WARN(d_logger,
                                    boost::format("LOST: File %s has missing portions - Last %i/%i, Current %i/%i")
                                        % temstr
                                        % EmwinFileIndex[i][1] % EmwinFileIndex[i][2]
                                        % seq % EmwinFileIndex[i][2]);
                        EmwinFileIndex[i][0] = 0; //mark available
                        EmwinFileIndex[i][3] = 0; //reset use counter
                        //goto doneblock;
                        DoneBlock=true;
                        break;
                    }
                    if (!DoneBlock){
                        EmwinFileIndex[i][1] = seq;
                        EmwinPtrTable[seq - 1][i] = EmwinProcPtr;
                        //goto doneblock;
                        DoneBlock=true;
                        break;
                    }
                }
            }
        }
    }
    //doneblock:
    EmwinStart = false;

    return;
}

void CGoesProcessor::EmwinDisplay(int indx)
{
    unsigned char *CurrentDataPtr,*Emwinstopptr;
    int MaxBlks, remain = 0, filesize = 0;
    char temstr[1000], filename[FILENAME_SIZE+1];
    char srch[3] = {0x0d, 0x0a, 0};
    int cmp = 3, i, lng;
    int length;
    int index;
    int Error;
    HANDLE Emfile;
    char tempstr[200], tempstring[200];
    EmwinTempBufferPtr = EmwinTempBuffer;
    MaxBlks = EmwinFileIndex[indx][2];
    filesize = 0;

    for (i = 0; i < MaxBlks; i++)
    {
        //printf("%d\n", &EmwinPtrTable[i][indx]);  //JR Debug to counter Heap corruption
        CurrentDataPtr = EmwinPtrTable[i][indx];
        if (!CurrentDataPtr) {
          GR_LOG_ERROR(d_logger,
                       boost::format("Null pointer in EmwinPtrTable[%d][%d]")
                       % i % indx);
          return;
        }
        if ((remain = EmwinBufEnd-CurrentDataPtr) >= 1024)
        {
            memcpy(EmwinTempBufferPtr, CurrentDataPtr, 1024);
            EmwinTempBufferPtr += 1024;
        }
        else
        {
            memcpy(EmwinTempBufferPtr, CurrentDataPtr, remain);
            EmwinTempBufferPtr += remain;
            CurrentDataPtr = EmwinData;
            memcpy(EmwinTempBufferPtr, CurrentDataPtr, 1024 - remain);
            EmwinTempBufferPtr += (1024 - remain);
        }

        filesize += 1024;
    }

    *EmwinTempBufferPtr = 0;
    Emwinstopptr = EmwinTempBufferPtr;
    EmwinTempBufferPtr = EmwinTempBuffer;
    memcpy(filename, EmwinFileName[indx], FILENAME_SIZE);
    filename[FILENAME_SIZE] = 0;

    if (boost::iequals(filename + 9, "TXT")) {
        cmp = 0;
    }
    if (boost::iequals(filename + 9, "JPG")) {
        cmp = 0;
    }
    if (boost::iequals(filename + 9, "GIF")) {
        cmp = 0;
    }
    //Esteban
    if (boost::iequals(filename + 9, "ZIS")) {
        cmp = 1;
        // This case is different because we rename
    }
    if (boost::iequals(filename + 9, "UID")) {
        cmp = 0;
    }

    // SAVE EVERY FILE, APPEND COUNTER AT THE END
    char alphabet[]="0123456789";

#ifndef DISABLE_WRITE

    switch (cmp)
    {
    case 0: {        // TXT File
        // create a scope oustide of which file is closed automagically
        {
            std::ostringstream makefilepath;
            makefilepath << emwindir << '/' << filename;

            std::string filepath(makefilepath.str());
            std::ofstream outfile(filepath, std::ios::binary);
            if (!outfile) {
                GR_LOG_ERROR(d_logger, boost::format("Could not write output file %s") % filepath);
            }
            else {
                outfile.write(reinterpret_cast<char*>(EmwinTempBuffer), filesize);
                GR_LOG_INFO(d_logger, "Extracted " + filepath);
            }
        }

        if (!boost::iequals(statestring, "ALL")) {
            if (!boost::iequals((const char*)filename + 6, statestring)) break;
        }
        while (EmwinTempBufferPtr < Emwinstopptr)
        {
            index = strcspn((const char*)EmwinTempBufferPtr, srch);
            if (index > 0)
            {
                *(EmwinTempBufferPtr + index) = 0;
                length = strlen((const char*)EmwinTempBufferPtr);
                if (length <= 90)
                {
                    sprintf(tempstring,"        ");
                    strcat(tempstring, (char*)EmwinTempBufferPtr);
                }
                else
                {
                    while (length > 0)
                    {
                      memcpy(temstr, EmwinTempBufferPtr, lng=std::min(90, length));
                        temstr[lng] = 0;
                        sprintf(tempstring, "        ");
                        strcat(tempstring, (char*)temstr);
                        EmwinTempBufferPtr += lng;
                        length -= lng;
                    }
                }
            }

            EmwinTempBufferPtr += index + 1;
        }

        break;
    }
    case 1: {        //      ZIS

        //------------------------------------------------------
        // SAVE EVERY FILE, APPEND COUNTER AT THE END
        // Esteban
        std::string filepath("EMWINtemp/");
        filepath += std::string(filename) + "P";
        // create a scope oustide of which file is closed automagically
        {
            std::ostringstream makefilepath;
            makefilepath << emwindir << '/' << filename << 'P';

            std::string filepath(makefilepath.str());
            std::ofstream outfile(filepath, std::ios::binary);
            if (!outfile) {
                GR_LOG_ERROR(d_logger, boost::format("Could not write output file %s") % filepath);
            }
            else {
                outfile.write(reinterpret_cast<char*>(FileBuffer), numLRITbytes);
                GR_LOG_INFO(d_logger, "Extracted " + filepath);
            }
        }
        if (!boost::iequals(statestring, "ALL")) {
            if (!boost::iequals((const char*)filename + 6, statestring)) break;
        }
        while (EmwinTempBufferPtr < Emwinstopptr)
        {
            index = strcspn((const char*)EmwinTempBufferPtr, srch);
            if (index > 0)
              {
                *(EmwinTempBufferPtr + index) = 0;
                length = strlen((const char*)EmwinTempBufferPtr);
                if (length <= 90)
                {
                    sprintf(tempstring,"        ");
                    strcat(tempstring, (char*)EmwinTempBufferPtr);
                }
                else
                {
                    while (length > 0)
                    {
                      memcpy(temstr, EmwinTempBufferPtr, lng=std::min(90, length));
                        temstr[lng] = 0;
                        sprintf(tempstring, "        ");
                        strcat(tempstring, (char*)temstr);
                        EmwinTempBufferPtr += lng;
                        length -= lng;
                    }
                }
            }

            EmwinTempBufferPtr += index + 1;
        }

        break;
    }
    default:
        GR_LOG_WARN(d_logger, boost::format("Non-Standard File Found: %s") % tempstr);
        break;
    }

#endif  // ENABLE / DISABLE write 2 file.

    return;
}

bool CGoesProcessor::RiceDecompress(int Riceflags, int PPB)
{
    // THis library has a memory leak. Memory increases by ~16 kb every other time you execute this
    //                                                            ----------------
    // SOLVED: By deleting the pointer before you return
    //
    //CRiceDecompression::CRiceDecompression    (       int       Mask,
      // int     BitsPerPixel,
       // int     PixelsPerBlock,
       // int     PixelsPerScanline,
        //int     ScanLinesPerPacket
   // )

     //Tried to fix library bug by not doing a rice = new CRiceDecompression and delete rice.  instead, I made a local copy only
     //CRiceDecompression  = rice(49, 8, 16, NC, 1).  When the function RiceDecompresS() exited, I let windows call the destructor
     //implicitly. This still caused the crash.  There was no way to gracefully delete this class without it crashing in certain cases.  this bug
     //is unavoidable unless you predict the crash ahead of time and never create the class in the first place.  Sometimes you will be unable
     //To "predict" the failure until you try to run Rice, and then Rice will crash no matter what you do.  Sometimes you will "predict" that
     //a packet is bad when it is really good, and you will not attempt to decompress real valid data! Without fixing rice, is there no
     //true solution to this problem.  There will always be cases (that have yet to be tested), when Rice breaks.

     //My observation is that if Rice is called and the decompression fails, you have a high likeihood of crashing on the current packet
     //Or on the next packets with the same parameters.  One method is to detect when Rice fails, save its configuration (NC = NC_failed) data,
     //and then avoid making the class when NC = NC_failed.

      //if(NC == NC_failed){
        //  rice_created = 0;
       //  return false;
   //  }

     //This method failed in the case where there were 5 sub-packets, first four were NC = 896 and the fifth was 895.
     //The first one was 896 and failed.  The next 3 were rejected because NC = NC_failed = 896.  The sub-packet
     //That caused the failure was 895 and it was not equal to 896.  Rice was called, but then caused the crash.

     //This method also failed because I found a case where NC =  1728 and it caused a crash on its first try, so there was no way to save
     //the configuration in order to avoid it during the next call.

     //The failure always seems to happen at NC = 896 or less, or is equal to 1728.  I can simply force NC to be greater than 2000 to call the class.
     // I am sure there will be some good packets with NC between 1728 and 2000.  I am sure there will be a case where a I packet with NC > 2000 will
     //cause a crash eventually.  So far, this is the best method!

    // if((NC <= 2000) | (abs(NC - NC_failed) < 2)){

          //printf("NC = %d, NC_failed = %d, Expected Rice Failure: Exiting Rice Decompression Call\n", NC, NC_failed);
         // rice_created = 0;
         // return false;
     //}
    //printf("NC = %d before Rice constructor \n",NC);
    //CRiceDecompression *Rice =  new CRiceDecompression(177, 8, 32, NC, 1);
    rice_created = 1;
    //printf("NC = %d after Rice constructor \n", NC);
    LRITimageptr = (unsigned char*)malloc(sizeof(unsigned char) * (NC+1));

    /////
    if (Riceflags == 0 || PPB == 0){
    	return false;
    }

////////

      SZ_com_t params;
	 int mask = Riceflags | SZ_RAW_OPTION_MASK; // By default NOAA dont as for RAW, but their images are RAW. No Compression header.

	  params.options_mask = mask;
	  params.bits_per_pixel = NB;
	  params.pixels_per_block = PPB;
	  params.pixels_per_scanline = NC;

	  size_t destLen = NC;
	  size_t outsize = NC;
	  int status = SZ_BufftoBuffDecompress(LRITimageptr, &outsize, reinterpret_cast<const char*>(riceptr), ricelength, &params);
	  if (status == SZ_OK){
		  RiceSize = NC*1;
		  NC_failed = 0;
		  return true;
	  }
	  else{
	      printf("NC = %d after Rice decompres() Call failed\n", NC);
	      NC_failed = NC;
	      return false;
	  }

    /*
    if (szip_uncompress_memory(SZ_RAW_OPTION_MASK | Riceflags, NB, PPB, NC, reinterpret_cast<const char*>(riceptr), ricelength, LRITimageptr, NC * 1) < 0 ) {
    //if (szip_uncompress_memory(SZ_RAW_OPTION_MASK | 49, 8, 16, NC, reinterpret_cast<const char*>(riceptr), ricelength, LRITimageptr, NC * 1) < 0 ) { // !rice->Decompress(riceptr, ricelength)){
        printf("NC = %d after Rice decompres() Call failed\n", NC);
    	printf("NC = %d \n", NC);
        NC_failed = NC;
        return false;
    }else{
        NC_failed = 0;
       // printf("NC = %d after Rice decompression: inide true \n", NC);
        //LRITimageptr = rice->Ptr();
        RiceSize = NC * 1; //rice->Size();
        //printf("NC = %d after Assigning sizes \n", NC);
        //delete rice;
        return true;
    }
*/
   // delete rice;
}

bool CGoesProcessor::CalcCRC(int filebytes)
{
    unsigned char *filepointer, *tempptr;
    unsigned short APID, Seq, Length, scounter;
    unsigned short Crc, CRCByte;
    unsigned char s1byte;
    unsigned short s2byte;
    int headlength,  totalheader = 1;
    int IMID, MaxSeg, MaxCol, MaxRow;
    int Riceflags, PPB, SLPP;
    int Param, Compress;
    char chardat[5000];
    int CFAC, LFAC, COFF, LOFF;
    numLRITbytes = 0;
    LRITfiletype = 130;
    filepointer = FileBuffer;
    bool rice_status;
    int filetype;

    while (lpBufPtr < LRITFile[APIDindex] + filebytes)
    {
                // See Source Packet Structure (TP_PDU)
        APID = *lpBufPtr;
        APID = ((APID<<8) | *(lpBufPtr+1)) & 0x7ff; //7FF=111 1111 1111
                // APID field is 11 bits. File identifier ranges from 0--> 2047
                // Example:
                // APID_0 = 0000 1001 --> (APID_0 <<8) = 0000 1001 0000 0000
                // APID_1 = 1000 0011 --> (APID_0 <<8) | (APID_1) & 0x7ff = 0000 0001 1000 0011
                //
                // Note that  lpBufPtr is not increased in the previous
        lpBufPtr += 2;
                // Increases by 2 bytes = 16 bits. It points at 2 bits from
                // Sequence Flags, and 6 bits from Packer Sequence Count
        Seq = *lpBufPtr;
        Seq = Seq>>6;

        scounter = (*lpBufPtr) & 0x37;  //x37= 11 111 --> 6 bits
        scounter = (scounter << 8) | *(lpBufPtr + 1);

        lpBufPtr += 2; // Pointing at the 1st byte from Packet Length
        Length = *lpBufPtr;
        Length = (Length << 8) | *(lpBufPtr + 1);
        Length++;
        //GR_LOG_CRIT(d_logger, "CalcCRC called");
        GR_LOG_INFO(d_debug_logger,
                    boost::format("APID = %i, Seq = %i, Count = %i, Length = %i") %
                    APID % Seq % scounter % Length);

        lpBufPtr += 2; // Points at the 1st byte from Application Data Field
        Crc = CRC(lpBufPtr, Length - 2);  // THe 2 last bytes are the CRC value
        CRCByte = *(lpBufPtr + Length - 2);
        CRCByte = (CRCByte << 8) | *(lpBufPtr + Length - 1);
        if (Crc != CRCByte)
            return false;

        // Rice pointer is at the begining of the Application Data Field
                riceptr = lpBufPtr;


                // Seq=1:  First segment of a file, extending to subsequent packets
                // Seq=3   One complete data file

                // Seq=0  Continuation packet, extending to subsequent
                // Seq=2  Last segment of a user data file
        if ((Seq == 1) || (Seq == 3))
        {
            if ((TMP_BUFFER_SIZE - (filepointer - FileBuffer)) < Length - 2){
                GR_LOG_ERROR(d_logger, "TMP Buffer Overload");
                return false;
            }
                        // Point tempptr to the end of the application data field
            tempptr = lpBufPtr + Length;
                        // Advance pointer by 10 bytes (Transport Layer Header: 2 + 8 bytes)
            lpBufPtr += 10;
                        // Copy the data portion to the file pointer location: TP _SDU
            memcpy(filepointer, lpBufPtr, Length-12); //Length - 12 because 2 bytes for CRC and 10 bytes for transport layer header
            numLRITbytes += Length - 12;
            filepointer += Length - 12;
            totalheader = 1;

            while (totalheader > 0)
            {
                                // Session Layer: Primary Header Record Type
                s1byte = *lpBufPtr;
                lpBufPtr++;
                s2byte = *lpBufPtr;
                headlength = (s2byte << 8) | *(lpBufPtr + 1);
                lpBufPtr += 2;
                //Debug
                int headtype = int(s1byte);
                GR_LOG_INFO(d_debug_logger,boost::format("Got header of type: %i") % headtype );

                switch (s1byte) //Primary Header in Session Layer S_PDU
                {
                case 0:  //Primary Header
                    LRITfiletype = Get(1);
                    totalheader = Get(4);
                    riceptr += totalheader;
                    datafieldlength = Get(8);
                    ricelength = totalheader;
                    totalheader = totalheader-headlength;

                    //Debug
                    filetype = int(LRITfiletype);
                    GR_LOG_INFO(d_debug_logger,boost::format("File type is: %i") % filetype );
                    break;
                case 1:  //Image Structure Record
                    NB = Get(1);
                    NC = Get(2);
                    NL = Get(2);
                    CFLG = Get(1);
                    GR_LOG_INFO(d_debug_logger,boost::format("BN is %i,  NC is %i, NL is %i, CFLG is %i") % int(NB)%int(NC)%int(NL)%int(CFLG) );
                    totalheader = totalheader - headlength;
                    break;
                case 2:  //Image Navigation Record
                    memcpy(chardat, lpBufPtr, 32);
                    chardat[32] = 0;
                    lpBufPtr += 32;
                    CFAC = Get(4);
                    LFAC = Get(4);
                    COFF = Get(4);
                    LOFF = Get(4);
                    totalheader = totalheader - headlength;
                    break;
                case 3:  //Image Data Function
                    if ((headlength - 3) > 5000)
                        return false;
                    memcpy(chardat, lpBufPtr, (headlength - 3));
                    chardat[(headlength - 3)] = 0;
                    lpBufPtr += (headlength - 3);
                    totalheader = totalheader - headlength;
                    break;
                case 4: //Annotation Record
                    if ((headlength - 3) > 5000)
                        return false;
                    memcpy(LRITfilename, lpBufPtr, (headlength - 3));
                    LRITfilename[(headlength - 3)] = 0;
                    LRIThavename = true;
                    lpBufPtr += (headlength - 3);
                    totalheader = totalheader - headlength;
                    break;
                case 5: //Time Stamp
                    lpBufPtr += 7;
                    totalheader = totalheader - headlength;
                    break;
                case 6: //Ancillary Text Record
                    if ((headlength - 3) > 5000)
                        return false;
                    memcpy(chardat, lpBufPtr, (headlength - 3));
                    lpBufPtr += (headlength - 3);
                    totalheader = totalheader - headlength;
                    break;

                case 128:  //Image Segment ID Record
                    IMID = Get(2);
                    SegNum = Get(2);
                    StartCol = Get(2);
                    StartLine = Get(2);
                    MaxSeg = Get(2);
                    MaxCol = Get(2);
                    MaxRow = Get(2);
                    totalheader = totalheader - headlength;
                    break;

                case 129: //NOAA LRIT Record
                                        // This includes DCS products
                    memcpy(chardat, lpBufPtr, 4);
                                        //void * memcpy ( void * destination, const void * source, size_t num );
                    chardat[4] = 0;
                    lpBufPtr += 4;
                    SPID = Get(2);
                                        /*
                                        if (int(SPID)==8){
                                                printf("Found DCS Packet\n");
                                        }else{
                                                printf("Product ID=%i\n",int(SPID));
                                        }
                                        */
                    SubPID = Get(2);
                    Param = Get(2);
                    Compress = Get(1);
                    {
                      std::string spidstr("UNKNOWN");
                      if (int(SPID)==8) spidstr = "DCS PRODUCT";
                      else if(int(SPID)==42) spidstr = "EMWIN";
                      else spidstr = "NOAA";
                      GR_LOG_INFO(d_debug_logger,
                                  boost::format("%s: PID = %i, SubPI = %i, Param = %i, Compress = %i")
                                      % spidstr % int(SPID) % int(SubPID) % int(Param) % int(Compress));
                    }
                    totalheader = totalheader - headlength;

                    break;
                case 130: //Header Structure Record
                    if ((headlength - 3) > 5000)
                        return false;
                    memcpy(chardat, lpBufPtr, (headlength - 3));
                    lpBufPtr += (headlength - 3);
                    totalheader = totalheader - headlength;
                    break;

                case 131:  //Rice Compression Record
                    Riceflags = Get(2);
                    PPB = Get(1);
                    SLPP = Get(1);
                    totalheader = totalheader - headlength;
                    break;
                case 132: //Data Filename
                    if ((headlength - 3) > 5000)
                        return false;
                    memcpy(chardat, lpBufPtr, (headlength - 3));
                    chardat[(headlength - 3)] = 0;
                    lpBufPtr += (headlength - 3);
                    totalheader = totalheader - headlength;
                    break;
                case 214: //EMWIN file
                    break;
                default:
                    GR_LOG_WARN(d_logger, boost::format("An unhandled header %i was found") % s1byte);
                    lpBufPtr += (headlength - 3);
                    totalheader = totalheader - headlength;
                    break;
                }
            }
                        // Seq=3   One complete data file
            if ((Seq == 3) && (LRITfiletype == 214)) //EMWIN file
            {
                if ((EmwinBufferLeft = (EmwinBufEnd - EmwinPtr)) >= 836)
                {
                    memcpy(EmwinPtr, lpBufPtr, 836);
                    EmwinPtr += 836;
                    EmwinBytes += 836;
                }
                else
                {
                    memcpy(EmwinPtr, lpBufPtr, EmwinBufferLeft);
                    EmwinPtr = EmwinData;
                    EmwinBytes += EmwinBufferLeft;
                    memcpy(EmwinPtr, lpBufPtr + EmwinBufferLeft, (836 - EmwinBufferLeft));
                    EmwinPtr += (836 - EmwinBufferLeft);
                    EmwinBytes += (836 - EmwinBufferLeft);
                }

                ProcessEMWINfile();
            }
            lpBufPtr = tempptr;
        }

        if ((Seq == 0) || (Seq == 2))
        {
            ricelength = Length - 2;
            if (LRITfiletype == 0 && int(Compress) == 1)
            {

                // printf("Before Calling Rice\n");
                rice_status = RiceDecompress(Riceflags, PPB);
               // printf("After Calling Rice\n");

              if (rice_status)
                {
                    //printf("NC = %d after returning from rice true \n", NC);
                    memcpy(filepointer, LRITimageptr, RiceSize);

                    free(LRITimageptr);
                    filepointer += RiceSize;
                    numLRITbytes += RiceSize;

                   // printf("After Checking rice status but before delete\n");
                    //if(rice_created) {}//delete rice;
                     //printf("After Checking rice status but after delete\n");
                }
              else{
                  printf("Rice Decompression Failed\n");
                 //  printf("After Checking rice status but before delete witin false\n");
              //  if(rice_created) {} //{delete rice;} //{printf("deleting rice\n"); delete rice;}
                   // printf("After Checking rice status but after delete witin false\n");
                    //return false;
              }
            }
            else
            {
                memcpy(filepointer, riceptr, ricelength);
                filepointer += ricelength;
                numLRITbytes += ricelength;
            }
            lpBufPtr += Length;
        }
    }
    return true;
}

unsigned short CGoesProcessor::CRC(const unsigned char* buf, int length)
{
    unsigned short Crc = 0xffff, ind = 0;

    while (length--)
        Crc = (Crc << 8) ^ crcTable[(Crc >> 8) ^ (unsigned short)buf[ind++]];

    return Crc;
}

int CGoesProcessor::Get(int bytes)
{
    int j, result = 0;

    for (j = 0; j < bytes; j++)
    {
        result = (result << 8) | *(lpBufPtr + j);
    }
    lpBufPtr += bytes;

    return result;
}
