from GOESLayer import GOESLayer
import sys
import os
from .frametypes import VCDUMsg
from .frametypes import MPDUMsg
from .opsmsgs import MPDUOpsMsg
import logging
import multiprocessing

##
# @brief MPDU Layer class.
# @details Third layer of LRIT packet processing. Extracts MPDU data from VCDU
# and performs some basic checks before passing it along.
# @author Miguel Dominguez
# @date 2016
class MPDULayer(GOESLayer):

    ##
    # @fn __init__
    # @brief Constructor.
    # @details Creates a logger and initializes VCDU counter
    def __init__(self):
        GOESLayer.__init__(self)
        self.logger = logging.getLogger(__name__)
        self.vcduCount = [None,] * (1 + VCDUMsg.VCID_MASK)
        #self.logger.setLevel(logging.DEBUG)
    ##
    # @fn processData
    # @brief Extracts MPDU data from VCDU and performs some basic checks.
    # @detail This performs a check for missing frames in any given Virtual
    # Channel. If so this is passed forward in an Ops Msg.  Also checks again
    # for inversion by checkign for impossible vaues of MPDU header information.
    # This may seem strange because we already check for inverted values in the
    # CADU layer, but there is a rare case I found where the preamble is not
    # inverted, but where the rest of the frame is. Weird!  Also, if the data
    # is fill, it is not passed forward.
    # @param[in] data VCDUMsg Object
    # @param[in] opsMsgs Information about fill or RS decoding failures
    # @param[out] outData MPDUMsg object
    # @param[out] outOpsMsg MPDUOpsMsg Object, mentioning missing frames or if
    # a new transport frame begins in this MPDU
    # @warning VCDU Count logic should be checked because it does not correctly
    # take into the account the case where the VCDU counter overflows and wraps
    # back around.
    def processData(self, data, opsMsg):
        self.logger.debug('Incoming VCDU')
        for line in str(data).split('\n'):
            self.logger.debug(line)

        output = MPDUMsg()
        outOpsMsg = MPDUOpsMsg()
        outOpsMsg.vcID = data.vcID

        #Assume data is a VCDUMsg, parse into MPDU
        output.fromFrame(data.data)

        self.logger.debug("Parsed MPDU")
        for line in str(output).split('\n'):
            self.logger.debug(line)

        # invalid first header pointer, is header inverted?
        if 884 < output.firstHeaderPointer < 2047:
            # inversion won't work for pointers in this region
            if output.firstHeaderPointer < 1163 and data.vcID != 63:
                self.logger.info('Invalid first header pointer - DISCARD - '
                                 'VCID=%d FHP=%d RSerrors=(%s)',
                                 outOpsMsg.vcID, output.firstHeaderPointer,
                                 ', '.join(str(x) for x in opsMsg.RSErrorLocations()))
                return None, outOpsMsg

            # otherwise invert and continue
            elif output.firstHeaderPointer >= 1163:
                self.logger.info('Invalid first header pointer - INVERT - '
                                 'VCID=%d, FHP=%d, RS errors=(%s)',
                                 outOpsMsg.vcID, output.firstHeaderPointer,
                                 ', '.join(str(x) for x in opsMsg.RSErrorLocations()))
                #output.firstHeaderPointer = 2048 - output.firstHeaderPointer
                #TODO:  Is this right? why would the data be inverted but not the header information?
                output.fromFrame([d ^ 0xFF for d in data.data])

        # easy way to default if entry is None
        lastCount = self.vcduCount[data.vcID] or (data.counter - 1) % data.MAX_VCDU_COUNTER
        self.vcduCount[data.vcID] = data.counter

        # Check if we have missed frames
        delta = data.counter - lastCount
        if delta < 0:
            #delta += data.MAX_VCDU_COUNTER
            delta = abs(delta)
        if delta > 1:
            outOpsMsg.setMissedFrames(delta)
            self.logger.warn('Missed %d frames - VCID=%d Expected=%d Received=%d',
                             outOpsMsg.getMissedFrames(), data.vcID,
                             (lastCount + 1) % data.MAX_VCDU_COUNTER, data.counter)
        elif delta == 0:
            self.logger.warn('Repeated frame count - VCID=%d Expected=%d Received=%d',
                             data.vcID, (lastCount + 1) % data.MAX_VCDU_COUNTER, data.counter)

        # Do not pass this data forward
        if data.vcID == VCDUMsg.FILL_VCID:
            data.calcBER()
            self.logger.info('Fill VCID - DISCARD')
            outOpsMsg.isFill = True
            return None, outOpsMsg

        outOpsMsg.setIsNewTransportFrame(0 <= output.firstHeaderPointer < MPDUMsg.NOT_NEW_FRAME)

        #Feed the Virtual Channel ID forward to CPPDU, which will actually use it
        if min(output.data) < 0:    # does this happen?
            raise ValueError

        return output, outOpsMsg
