#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Copyright 2016 The Aerospace Corporation.
# 
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
# 
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
# 

import os
import pyface.api as pf
import traits.api as tr
import traitsui.api as trui
import traitsui.file_dialog  as trfd
from pyface.qt import QtGui, QtCore
from traitsui.qt4.editor import Editor as Qt4Editor
from traitsui.image.image import ImageLibrary

import logging
logger=logging.getLogger(__name__)

# To embed a non-TraitsUI element (widget/dialog/etc) a custom editor_factory
# must be declared with the below argument list (parent, editor) where the
# editor is a handle to the class that is including the element. The return
# item must be a PyQt element (probably derived from the Widget base class).
# Example where 'dialog' is a PyQt object (presumably derived from QWidget)
# traits_view = trui.View( trui.UCustom('dialog',
#                                       editor=trui.CustomEditor(qt_to_traitsui_factory)),
#                    resizable=True)
def qt_to_traitsui_factory(parent, editor):
    return getattr(editor.object, editor.name)

# Similar to above, embed the factory with lambda and just make this a function call
# that returns the desired custom editor.
# e.g. view=trui.View(trui.UCustom('QDialog_object', editor=EmbeddedPyQtEditor()))
def get_qt_editor():
    return trui.CustomEditor(qt_to_traitsui_factory)

from PyQt4 import Qt
class QWidgetWrapper(tr.HasTraits):
    """Wraps a QWidget instance in another view s.t. you can update the widget."""
    qwidget = tr.Instance(Qt.QWidget)
    
    view = trui.View(trui.UItem('qwidget', editor=get_qt_editor()))

    # the following get/set methods set up a sort of transparent object proxy
    def __getattribute__(self, name):
        try:
            # works only for local attributes (e.g. qwidget, view)
            return object.__getattribute__(self, name)
        except AttributeError:
            # fallback to trying qwidget
            return object.__getattribute__(self.qwidget, name)
        
    def __setattribute__(self, name, value):
        try:
            object.__setattribute__(self, name, value)
        except AttributeError:
            object.__setattribute__(self.qwidget, name, value)

class ModalTraitsDialog(pf.Dialog):
    ''' Convenience class to create a modal dialog for an existing traits view. '''
    def __init__(self, traits_view, **kw):
        super(ModalTraitsDialog, self).__init__(**kw)
        self.traits_view=traits_view
        
    def _create_dialog_area(self, parent):
        return self.traits_view.edit_traits().control

class _ToggleButtonEditor(Qt4Editor):
    """ Qt implementation of a toggle button.
    """
    checked_icon = tr.Instance(QtGui.QIcon)
    unchecked_icon = tr.Instance(QtGui.QIcon)
    checked_label = tr.Unicode()
    unchecked_label = tr.Unicode()
    
    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.checked_label = self.factory.checked_label
        self.unchecked_label = self.factory.unchecked_label
        if self.factory.checked_image_name != '':
            image=ImageLibrary.image_resource(self.factory.checked_image_name)
            self.checked_icon=image.create_icon(self.factory.image_size)
        if self.factory.unchecked_image_name != '':
            image=ImageLibrary.image_resource(self.factory.unchecked_image_name)
            self.unchecked_icon=image.create_icon(self.factory.image_size)
        self.control = QtGui.QPushButton()
        self.control.setCheckable(True)
        self.control.setChecked(bool(self.value))
        self.control.clicked.connect(self.update_object)        
        self.update_control()
        self.set_tooltip()

    def update_control(self):
        if self.control == None: return
        if self.value:
            if self.checked_icon != None:
                self.control.setIcon(self.checked_icon)
            self.control.setText(self.string_value(self.checked_label))
        else:
            if self.unchecked_icon != None:
                self.control.setIcon(self.unchecked_icon)
            self.control.setText(self.string_value(self.unchecked_label))                
    
    def update_object(self):
        """ Updates the object when the button is clicked.
        """
        self.value = self.control.isChecked()
        self.update_control()

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.control.setChecked(bool(self.value))
        self.update_control()


class ToggleButtonEditor(trui.EditorFactory):
    """ A toggle button for editing Bool traits.
    """

    # The button label
    label = tr.Unicode()
    unchecked_label = tr.Unicode()
    checked_label = tr.Unicode()
    checked_image_name = tr.Str
    unchecked_image_name = tr.Str
    image_size = tr.Int(16)
    def __init__(self, **kws):
        super(ToggleButtonEditor,self).__init__(**kws)
        if self.unchecked_label == '':
            self.unchecked_label = self.label
        if self.checked_label == '':
            self.checked_label = self.label


    def _get_simple_editor_class(self):
        return _ToggleButtonEditor

class _AnimatedButtonEditor(Qt4Editor):
    """ Qt implementation of a toggle button.
    """
    movie = tr.Instance(QtGui.QMovie)
    checked_label = tr.Unicode()
    unchecked_label = tr.Unicode()
    _status = tr.Str
    _status_obj = tr.Instance(tr.HasTraits)
    
    def init(self, parent):         
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.checked_label = self.factory.checked_label
        self.unchecked_label = self.factory.unchecked_label
        self.movie = QtGui.QMovie(self.factory.gif)
        self.movie.setScaledSize(QtCore.QSize(self.factory.gif_width, self.factory.gif_height))
        self.control = QtGui.QPushButton()
        # Cannot use setMovie on a QPushButton but can use a movie with a signal
        # to set the Icon from:
        # http://stackoverflow.com/questions/15374191/how-to-set-animated-icon-to-qpushbutton-in-qt5
        QtCore.QObject.connect(self.movie, QtCore.SIGNAL('frameChanged(int)'),self.set_icon)
        self.control.setCheckable(True)
        self.control.setChecked(self.value)
        self.control.clicked.connect(self.update_object)
        # Start/stop the movie to get a valid Pixmap for the initial icon
        self.movie.start()
        self.movie.stop()
        self.set_tooltip()
        # If status is defined, set up a notification that it has changed to update the movie
        if self.factory.status != '':
            if len(self.factory.status.split('.'))>2:
                raise ex.NotImplementedError, 'AnimatedButtonEditor does not support "status" parameters not support extended traits names yet'
            if self.factory.status.startswith('controller.'):
                self._status = self.factory.status.split('controller.',1)[1] # FIXME - make cleaner than using split for 'replace at beginning only'
                self._status_obj = self.object
            else:
                if self.factory.status.startswith('object.'):
                    self._status = self.factory.status.split('object.',1)[1] # FIXME - make cleaner than using split for 'replace at beginning only'
                else:
                    self._status = self.factory.status # no prepended controller/object
                # if controller, relative to controller.model, otherwise, relative to self
                if isinstance(self.object, trui.Controller):
                    self._status_obj = self.object.model
                else:
                    self._status_obj = self.object

            self._status_obj.on_trait_change(self.status_update, self._status)                
        self.update_control()
    
    def set_icon(self, frame):
        self.control.setIcon(QtGui.QIcon(self.movie.currentPixmap()))
        
    def update_control(self):
        if self.control == None: return
        # Update button text
        if self.value:
            self.control.setText(self.string_value(self.checked_label))
        else:
            self.control.setText(self.string_value(self.unchecked_label))
        # if using self.value for animation status call it
        # otherwise, it'll be caught with trait notification
        if self.factory.status == '':
            if self.value:
                self.movie.start()
            else:
                self.movie.stop()                

    def status_update(self):
        # This only applies if status != '' (i.e. it isn't self.value)
        # retrieve status value
        active = bool(getattr(self._status_obj, self._status))
        if active:
            self.movie.start()
        else:
            self.movie.stop()
        self.value = active
        
    def update_object(self):
        """ Updates the object when the button is clicked.
        """
        self.value = self.control.isChecked()
        self.update_control()
 
    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.control.setChecked(self.value)
        self.update_control()
 
 
class AnimatedButtonEditor(trui.EditorFactory):
    """ A toggle button for editing Bool traits.
    """
 
    # The button label
    label = tr.Unicode()
    unchecked_label = tr.Unicode()
    checked_label = tr.Unicode()
    status = tr.Str(desc='Extended trait name of object to use for active or not (will be treated as a Bool)')
    gif = tr.Str
    gif_width = tr.Int(16)
    gif_height = tr.Int(16)
     
    def __init__(self, **kws):
        super(AnimatedButtonEditor,self).__init__(**kws)
        if self.unchecked_label == '':
            self.unchecked_label = self.label
        if self.checked_label == '':
            self.checked_label = self.label
        if not os.path.exists(self.gif):
            if self.gif != '':
                logger.error('Cannot find %s for AnimatedButtonEditor'%self.gif)
            cur_dir = os.path.dirname(__file__)
            # We live in aero_py\gui_modules\ so go up one and back down to images
            self.gif = cur_dir+r'\..\images\busy_blue_spinner.gif'
 
    def _get_simple_editor_class(self):
        return _AnimatedButtonEditor

class _BoolImageEditor(Qt4Editor):
    """ Qt implementation of a toggle button.
    """
    true_icon = tr.Instance(QtGui.QIcon)
    false_icon = tr.Instance(QtGui.QIcon)

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """        
        image=ImageLibrary.image_resource(self.factory.true_image)
        self.true_icon=image.create_icon(self.factory.image_size)
        image=ImageLibrary.image_resource(self.factory.false_image)
        self.false_icon=image.create_icon(self.factory.image_size)
        self.control = QtGui.QLabel()
        self.update_image()
        self.set_tooltip()

    def update_image(self):
        if self.control == None: return
        size = QtCore.QSize(self.factory.image_size,self.factory.image_size)
        # TODO - look into QIcon vs QPixmap (esp for above ToggleButtonEditor)
        if self.value:
            self.control.setPixmap(self.true_icon.pixmap(size))
        else:
            self.control.setPixmap(self.false_icon.pixmap(size))
        
    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.update_image()

class BoolImageEditor(trui.EditorFactory):
    """ A toggle image for Bool traits.
    """
    true_image = tr.Str('@aero_images:green-on-16.png')
    false_image = tr.Str('@aero_images:green-off-16.png')
    image_size = tr.Int(16)

    def _get_simple_editor_class(self):
        return _BoolImageEditor

class _BoolAnimatedEditor(Qt4Editor):
    """ Qt implementation of an animated GIF.
    """
    movie = tr.Instance(QtGui.QMovie)
    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """        
        #self.movie = QtGui.QMovie(self.factory.gif)
        if not os.path.exists(self.factory.gif):
            # FIXME - using relative path for default, assumes aero_py/images/*.gif and
            # aero_py/gui_modules/misc.py stay the same as far as pathing goes
            base_dir = __file__.rsplit('\\',2)[0]
            self.factory.gif = base_dir +r'\images\busy_blue_flower.gif'
        self.movie = QtGui.QMovie(self.factory.gif)
        self.movie.setScaledSize(QtCore.QSize(self.factory.width,self.factory.height))
        self.control = QtGui.QLabel()
        self.control.setMovie(self.movie)
        self.movie.start()        
        self.set_tooltip()
        #self.update_image()

    def update_image(self):
        if self.control == None: return
        if self.value:
            self.movie.start()
        else:
            self.movie.stop()
        
    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.update_image()

class BoolAnimatedEditor(trui.EditorFactory):
    """ An animated gif for Bool traits.
    """
    gif = tr.Str()

    width = tr.Int(16)
    height = tr.Int(16)

    def _get_simple_editor_class(self):
        return _BoolAnimatedEditor

def test_toggle_button_editor():
    import aero_py.images
    
    class Demo(tr.HasTraits):
        toggle_state = tr.Bool(False)
        toggle_state2 = tr.Bool(False)
        
        traits_view = trui.View(
            trui.UItem('toggle_state', editor=ToggleButtonEditor(label='Toggle Me',checked_image_name='@aero_images:red-on-16.png',
                                                                              unchecked_image_name='@aero_images:red-off-16.png')),
            trui.UItem('toggle_state2', editor=ToggleButtonEditor(unchecked_label=u'Toggle Me Too!',
                                                             checked_label=u'Turn me off!')),
            trui.UItem('toggle_state2', editor=AnimatedButtonEditor(unchecked_label=u'Toggle Me Too!',
                                                             checked_label=u'Turn me off!',
                                                             status='toggle_state')),
            trui.UItem('toggle_state2', editor=BoolImageEditor()),
            trui.UItem('toggle_state2', editor=BoolAnimatedEditor()),
            buttons=['OK'],
            resizable=True,
            width=150,
        )
 
 
        def _toggle_state_changed(self, new):
            print 'State toggled to {}'.format(new)
 
    d = Demo()
    d.configure_traits()


class DictView(tr.HasTraits):
    obj = tr.Instance( object )

    def __init__(self, obj, **traits):
        super(DictView, self).__init__(**traits)
        self.obj = obj

    def trait_view(self, name=None, view_elements=None):
        return trui.View(
          trui.VGroup(
            trui.Item( 'obj',
                  label      = 'Debug',
                  id         = 'debug',
                  editor     = trui.ValueEditor(),
                  style      = 'custom',
                  dock       = 'horizontal',
                  show_label = False
            ),
          ),
          title     = 'Dictionary Viewer',
          width     = 800,
          height    = 600,
          resizable = True,
        )

# FIXME - redo this to rewrite/override the trfd.open_file/OpenFileDialog to have recent files list
class FileDialog (tr.HasTraits):
    filename = tr.File
    open = tr.Button('Open...')
    filter = tr.Str('Python file (*.py)|*.py|All files (*.*)|*.*')
    entries = tr.Int(10, desc='Number of recent entries to remember')
    recent_files = tr.List(tr.Str)
    id = tr.Str(desc='ID used for traits persistence, unsure how this works, leaving for potential future use')
    view = trui.View(
        trui.HGroup(
            trui.Item( 'open', show_label = False ),
            '_',
            trui.Item( 'filename', style = 'readonly', springy = True )
        ),
        width = 0.5
    )

    def _open_changed ( self ):
        filename = trfd.open_file( extensions = trfd.TextInfo(),
                               filter     = self.filter,
                               id         = self.id )
        if filename != '':
            self.filename = filename
            # Remove filename if already in recent
            if filename in self.recent_files:
                self.recent_files.remove(filename)
            self.recent_files.insert(0, filename) # put filename at top of list
            


class ProgressBar(tr.HasTraits):
    bar = tr.Instance(QtGui.QProgressBar)
    min = tr.Float(0)
    max = tr.Float(1)
    _range = tr.Float(1)
    current = tr.Float(0)
    show_percent = tr.Bool(True)
    orientation = tr.Enum('Horizontal','Vertical')
    
    view = trui.View(trui.UCustom('bar',
                                  # used lambda to return the bar object for the factory function
                                  # by getting the attribute name ('bar' in this case) from the
                                  # editor object (self in this case)
                                  editor=trui.CustomEditor(lambda p,e: getattr(e.object,e.name))),
                     resizable=True
                     )

    def __init__(self, *args, **kws):
        super(ProgressBar, self).__init__(*args, **kws)
        self.bar = QtGui.QProgressBar()
        self.bar.setMinimum(0) # always zero
        self.bar.setMaximum(10000) # always 10000 to display as percent with 2 signif digits
        self._orientation_changed()
        self._show_percent_changed()
    
    @tr.on_trait_change('current')
    def _update(self):
        value = int((self.current-self.min)/self._range*10000) # essentially percent with 2 extra digits
        self.bar.setValue(value)
        QtGui.QApplication.processEvents()
    
    @tr.on_trait_change('min,max')
    def _update_range(self):
        self._range = self.max-self.min
        
    def _show_percent_changed(self):
        self.bar.setTextVisible(self.show_percent)
    
    def _orientation_changed(self):
        self.bar.setOrientation(QtCore.Qt.Horizontal if self.orientation == 'Horizontal'
                                else QtCore.Qt.Vertical)


class Slider(tr.HasTraits):
    slider = tr.Instance(QtGui.QSlider)
    label=tr.Str
    current = tr.Float
    min = tr.Float(0)
    max = tr.Float(100)
    show_percent = tr.Bool(True)
    orientation = tr.Enum('Horizontal','Vertical')
    
    view = trui.View(trui.VGroup(trui.HGroup(
                                 trui.UCustom('slider',
                                  # used lambda to return the bar object for the factory function
                                  # by getting the attribute name ('bar' in this case) from the
                                  # editor object (self in this case)
                                  editor=trui.CustomEditor(lambda p,e: getattr(e.object,e.name))),
                                 trui.UReadonly('label', visible_when='show_percent==True and orientation=="Horizontal"', width=-50),
                                 ),
                                 trui.UReadonly('label', visible_when='show_percent==True and orientation=="Vertical"', width=-50),
                            ),
                     resizable=True
                     )

    def __init__(self, *args, **kws):
        super(Slider, self).__init__(*args, **kws)
        self.slider = QtGui.QSlider()
        self.slider.valueChanged[int].connect(self.update)
        self._orientation_changed()
        self._max_changed()
        self._min_changed()
    
    def _current_changed(self):
        self.update(self.current)
        
    def update(self, value):
        if value<self.min:
            value = self.min
        if value>self.max:
            value = self.max
        self.slider.setValue(value)
        percent = (float(value)-self.min)/(self.max-self.min)*100
        self.label='%3.1f%%'%percent
        QtGui.QApplication.processEvents()
        self.current = value
    
    def _max_changed(self):
        if self.slider!=None:
            self.slider.setMaximum(self.max)

    def _min_changed(self):
        if self.slider!=None:
            self.slider.setMinimum(self.min)
    
    def _orientation_changed(self):
        self.slider.setOrientation(QtCore.Qt.Horizontal if self.orientation == 'Horizontal'
                                else QtCore.Qt.Vertical)
if __name__=='__main__':
    # This file lives under ../aero_py/gui_modules/ so ensure 2 directories up is in the system path
    import sys, os
    fname=os.path.realpath(__file__)
    # Get the directory above aero_gui by removing everything after 3 \'s from the right
    py_base = fname.rsplit('\\',3)[0] # Use '\\' to get single backslash
    if py_base not in sys.path:
        sys.path.append(py_base)    
    #f=FileDialog()
    #f.configure_traits()
    test_toggle_button_editor()
    #gui=ProgressBar()
#     gui=Slider()
#     gui.update(35)
#     gui.show_percent=True
#     gui.orientation = 'Horizontal'
#     gui.orientation = 'Vertical'
#     gui.configure_traits()
